import pytest

from metadata_crawler.api.storage_backend import TemplateMixin


def test_basic_string() -> None:
    assert (
        TemplateMixin().render_templates("Hello {{ name }}!", {"name": "Ada"})
        == "Hello Ada!"
    )


def test_nested_containers_and_env(monkeypatch: pytest.MonkeyPatch) -> None:
    # Prepare environment variables used by ENV / env() / getenv
    monkeypatch.setenv("HOME", "/home/ada")
    monkeypatch.delenv("PORT", raising=False)  # ensure default path is used

    data = {
        "greet": "Hi {{ name }} from {{ ENV.HOME }}",
        "port": "{{ env('PORT', '8000') }}",
        "items": [
            "{{ count }} thing(s)",
            42,
            ("x={{ x }}", "{{ 'HOME'|getenv('/tmp') }}"),
        ],
        "keys": {"k{{ n }}": "v{{ n }}"},
        "set": {"{{ name }}", 1},
    }
    ctx = {"name": "Ada", "count": 3, "x": 7, "n": 1}

    out = TemplateMixin().render_templates(data, ctx)

    assert "Hi Ada" in out["greet"]
    assert out["port"] == "8000"
    assert out["items"][0] == "3 thing(s)"
    assert out["items"][2][0] == "x=7"
    assert out["keys"] == {"k1": "v1"}
    assert out["set"] == {"Ada", 1}


def test_multi_pass_expansion():
    # a -> {{ b }} -> "done" with two passes
    s = "{{ a }}"
    ctx = {"a": "{{ b }}", "b": "done"}

    assert TemplateMixin().render_templates(s, ctx, max_passes=2) == "done"
    assert TemplateMixin().render_templates(s, ctx, max_passes=1) == "{{ b }}"


def test_dict_key_collision_last_wins():
    data = {"a{{ n }}": 1, "a{{ m }}": 2}
    ctx = {"n": 1, "m": 1}
    assert TemplateMixin().render_templates(data, ctx) == {"a1": 2}


def test_non_string_scalars_unchanged():
    assert TemplateMixin().render_templates(123, {}) == 123
    assert TemplateMixin().render_templates(3.14, {}) == 3.14
    assert TemplateMixin().render_templates(True, {}) is True
    assert TemplateMixin().render_templates(None, {}) is None



"""Collection of aync data ingest classes."""

from __future__ import annotations

import asyncio
import logging
import os
import time
from concurrent.futures import ThreadPoolExecutor
from types import TracebackType
from typing import Annotated, Any, Dict, List, Optional, Tuple, Type, cast

import aiohttp
import orjson

from ..api.cli import cli_function, cli_parameter
from ..api.index import BaseIndex
from ..api.metadata_stores import IndexName
from ..logger import logger


class SolrIndex(BaseIndex):
    """Ingest metadata into an apache solr server."""

    senteniel: Optional[bytes] = None

    def __post_init__(self) -> None:
        self.timeout = aiohttp.ClientTimeout(
            connect=10, sock_connect=10, sock_read=180, total=None
        )
        self.semaphore = asyncio.Event()
        self.max_http_workers: int = 0
        queue_max: int = 128
        encode_workers: int = 4
        self._uri: str = ""
        self.cpu_pool = ThreadPoolExecutor(max_workers=encode_workers)
        self.producer_queue: asyncio.Queue[Tuple[str, Optional[bytes]]] = (
            asyncio.Queue(maxsize=queue_max)
        )
        self.connector = aiohttp.TCPConnector(
            ttl_dns_cache=300,
            use_dns_cache=True,
            enable_cleanup_closed=True,
        )

    async def solr_url(self, server: str, core: str) -> str:
        """Construct the solr url from a given solr core."""
        if not self._uri:
            scheme, _, server = server.rpartition("://")
            scheme = scheme or "http"
            solr_server, _, solr_port = server.partition(":")
            solr_server = solr_server or "localhost"
            uri = f"{scheme}://{solr_server}"
            uri = f"{uri}:{solr_port}" if solr_port else uri
            self._uri = f"{uri}/solr"
        return f"{self._uri}/{core}/update/json?commit=true"

    @cli_function(
        help="Remove metadata from the apache solr server.",
    )
    async def delete(
        self,
        *,
        server: Annotated[
            Optional[str],
            cli_parameter(
                "-sv",
                "--server",
                help="The <host>:<port> to the solr server",
                type=str,
            ),
        ] = None,
        facets: Annotated[
            Optional[List[tuple[str, str]]],
            cli_parameter(
                "-f",
                "--facets",
                type=str,
                nargs=2,
                action="append",
                help="Search facets matching the delete query.",
            ),
        ] = None,
        latest_version: Annotated[
            str,
            cli_parameter(
                "--latest-version",
                type=str,
                help="Name of the core holding 'latest' metadata.",
            ),
        ] = IndexName().latest,
        all_versions: Annotated[
            str,
            cli_parameter(
                "--all-versions",
                type=str,
                help="Name of the core holding 'all' metadata versions.",
            ),
        ] = IndexName().all,
    ) -> None:
        """Remove metadata from the apache solr server."""
        query = []
        for key, value in facets or []:
            if key.lower() == "file":
                if value[0] in (os.sep, "/"):
                    value = f"\\{value}"
                value = value.replace(":", "\\:")
            else:
                value = value.lower()
            query.append(f"{key.lower()}:{value}")
        query_str = " AND ".join(query)
        server = server or ""
        async with aiohttp.ClientSession(timeout=self.timeout) as session:
            logger.debug("Deleting entries matching %s", query_str)
            for core in (all_versions, latest_version):
                url = await self.solr_url(server, core)
                async with session.post(
                    url, json={"delete": {"query": query_str}}
                ) as resp:
                    level = (
                        logging.WARNING
                        if resp.status not in (200, 201)
                        else logging.DEBUG
                    )
                    logger.log(level, await resp.text())

    def _convert(self, metadata: Dict[str, Any]) -> Dict[str, Any]:
        for k, v in metadata.items():
            match self.index_schema[k].type:
                case "bbox":
                    metadata[k] = f"ENVELOPE({v[0]}, {v[1]}, {v[3]}, {v[2]})"
                case "daterange":
                    metadata[k] = f"[{v[0].isoformat()} TO {v[-1].isoformat()}]"

        return metadata

    def _encode_payload(self, chunk: List[Dict[str, Any]]) -> bytes:
        """CPU-bound: convert docs and JSON-encode off the event loop."""
        return orjson.dumps([self._convert(x) for x in chunk])

    async def _post_chunk(
        self,
        session: aiohttp.ClientSession,
        url: str,
        body: bytes,
    ) -> None:
        """POST one batch with minimal overhead and simple retries."""
        status = 500
        t0 = time.perf_counter()
        try:
            async with session.post(
                url, data=body, headers={"Content-Type": "application/json"}
            ) as resp:
                status = resp.status
                await resp.read()

        except Exception as error:
            logger.log(
                logging.WARNING,
                error,
                exc_info=logger.level < logging.INFO,
            )
            return
        logger.debug(
            "POST %s -> %i (index time: %.3f)",
            url,
            status,
            time.perf_counter() - t0,
        )

    async def consumer(self, session: aiohttp.ClientSession) -> None:
        """Consume the metadata read by the porducers."""
        while True:
            update_url, body = await self.producer_queue.get()
            if body is self.senteniel:
                self.producer_queue.task_done()
                break
            try:
                await self._post_chunk(session, update_url, cast(bytes, body))
            finally:
                self.producer_queue.task_done()

    async def _index_core(
        self,
        session: aiohttp.ClientSession,
        server: str,
        core: str,
        suffix: str,
        http_workers: int = 8,
    ) -> None:
        """Zero-copy-ish, backpressured, bounded-concurrency indexer.

        - No per-batch commit.
        - Bounded queue so tasks don't pile up.
        - Constant number of worker tasks (not O(batches)).
        """
        base_url = await self.solr_url(server, core + suffix)
        update_url = base_url.split("?", 1)[0]  # guard
        loop = asyncio.get_running_loop()
        async for batch in self.get_metadata(core):
            body = await loop.run_in_executor(
                self.cpu_pool, self._encode_payload, batch
            )
            await self.producer_queue.put((update_url, body))
        commit_url = f"{update_url}?commit=true"
        async with session.post(
            commit_url,
            data=b"[]",
            headers={"Content-Type": "application/json"},
        ) as resp:
            if resp.status >= 400:
                text = await resp.text()
                logger.warning(
                    "COMMIT %s -> %i: %s", commit_url, resp.status, text
                )

    async def __aexit__(
        self,
        exc_type: Optional[Type[BaseException]],
        exc_val: Optional[BaseException],
        exc_tb: Optional[TracebackType],
    ) -> None:

        try:
            self.producer_queue.shutdown()
        except AttributeError:  # pragma: no cover
            pass  # prgama: no cover
        self.cpu_pool.shutdown()

    @cli_function(
        help="Add metadata to the apache solr metadata server.",
    )
    async def index(
        self,
        *,
        server: Annotated[
            Optional[str],
            cli_parameter(
                "-sv",
                "--server",
                help="The <host>:<port> to the solr server",
                type=str,
            ),
        ] = None,
        index_suffix: Annotated[
            Optional[str],
            cli_parameter(
                "--index-suffix",
                help="Suffix for the latest and all version collections.",
                type=str,
            ),
        ] = None,
        http_workers: Annotated[
            int,
            cli_parameter(
                "--http-workers", help="Number of ingestion threads.", type=int
            ),
        ] = 8,
    ) -> None:
        """Add metadata to the apache solr metadata server."""
        async with aiohttp.ClientSession(
            timeout=self.timeout, connector=self.connector, raise_for_status=True
        ) as session:
            consumers = [
                asyncio.create_task(self.consumer(session))
                for _ in range(http_workers)
            ]
            async with asyncio.TaskGroup() as tg:
                for core in self.index_names:
                    tg.create_task(
                        self._index_core(
                            session,
                            server or "",
                            core,
                            suffix=index_suffix or "",
                            http_workers=http_workers,
                        )
                    )
            for _ in range(http_workers):
                await self.producer_queue.put(("", self.senteniel))
            await self.producer_queue.join()
            await asyncio.gather(*consumers)



"""Gather metadata and for adding them to a temporary metadata store."""

from __future__ import annotations

import asyncio
import os
from multiprocessing import Event, Value
from pathlib import Path
from types import TracebackType
from typing import (
    Any,
    AsyncIterator,
    Callable,
    Coroutine,
    Dict,
    Iterator,
    Optional,
    Tuple,
    Type,
    Union,
    cast,
)

import tomlkit

from .api.config import CrawlerSettings, DRSConfig
from .api.metadata_stores import CatalogueWriter, IndexName
from .api.storage_backend import PathTemplate
from .logger import logger
from .utils import (
    Counter,
    MetadataCrawlerException,
    create_async_iterator,
    print_performance,
)

ScanItem = Tuple[str, str, bool, bool]


class DataCollector:
    """Collect file objects from a given directory object and search for files.

    Parameters
    ----------
    config:
        Metadata-crawler config module.
    *search_objects:
        Paths of the search directories. e.g. `root_path` attr in drs_config
    uri: str
        the uir of the metadata store.
    password: str
        Password for the ingestion
    batch_size: int
        Batch size for the ingestion
    """

    def __init__(
        self,
        config: DRSConfig,
        metadata_store: Optional[
            Union[Path, str, Dict[str, Any], tomlkit.TOMLDocument]
        ],
        index_name: IndexName,
        *search_objects: CrawlerSettings,
        **kwargs: Any,
    ):
        self._search_objects = search_objects
        if not search_objects:
            raise MetadataCrawlerException("You have to give search directories")
        self._num_files: Counter = Value("i", 0)
        self.index_name = index_name
        self.config = config
        kwargs.setdefault("scan_concurrency", os.getenv("SCAN_CONCURRENCY", "64"))
        self._scan_concurrency: int = int(kwargs.pop("scan_concurrency", 64))
        self._scan_queue: asyncio.Queue[Optional[ScanItem]] = asyncio.Queue(
            maxsize=int(kwargs.pop("scan_queue_size", 10_000))
        )
        self._print_status = Event()
        self.ingest_queue = CatalogueWriter(
            str(metadata_store or "metadata.yaml"),
            index_name=index_name,
            config=self.config,
            **kwargs,
        )
        self.ingest_queue.run_consumer()
        self._max_files = int(cast(str, os.getenv("MDC_MAX_FILES", "-1")))

    @property
    def crawled_files(
        self,
    ) -> int:
        """Get the total number of crawled files."""
        return self._num_files.value

    @property
    def ingested_objects(self) -> int:
        """Get the number of ingested objects."""
        return self.ingest_queue.ingested_objects

    @property
    def search_objects(self) -> Iterator[tuple[str, str]]:
        """Async iterator for the search directories."""
        for cfg in self._search_objects:
            yield cfg.name, str(cfg.search_path)

    async def __aenter__(self) -> "DataCollector":
        return self

    async def __aexit__(
        self,
        exc_type: Optional[Type[BaseException]],
        exc: Optional[BaseException],
        tb: TracebackType,
    ) -> None:
        self._print_status.clear()
        self.ingest_queue.join_all_tasks()
        await self.ingest_queue.close()

        async def _safe_close(b: PathTemplate) -> None:
            try:
                await asyncio.wait_for(b.close(), timeout=3)
            except Exception:
                pass

        await asyncio.gather(
            *[_safe_close(ds.backend) for ds in self.config.datasets.values()],
            return_exceptions=True,
        )

    def _test_env(self) -> bool:
        return (
            True
            if self._max_files > 0 and self._max_files < self.crawled_files
            else False
        )

    async def _ingest_dir(
        self,
        drs_type: str,
        search_dir: str,
        iterable: bool = True,
        is_versioned: bool = True,
    ) -> None:
        silent = self.ingest_queue.silent
        if iterable:
            try:
                _sub_dirs = self.config.datasets[drs_type].backend.iterdir(
                    search_dir
                )
            except Exception as error:
                logger.error(error)
                return
        else:
            _sub_dirs = cast(
                AsyncIterator[str], create_async_iterator([search_dir])
            )
        rank = 0
        sub_dirs = []
        async for _dir in _sub_dirs:
            sub_dirs.append(_dir)
        sub_dirs.sort(reverse=True)
        for _dir in sub_dirs:
            async for _inp in self.config.datasets[drs_type].backend.rglob(
                _dir, self.config.datasets[drs_type].glob_pattern
            ):
                if self._test_env():
                    return
                await self.ingest_queue.put(
                    _inp, drs_type, name=self.index_name.all
                )
                if rank == 0 or is_versioned is False:
                    await self.ingest_queue.put(
                        _inp, drs_type, name=self.index_name.latest
                    )
                if silent is False:
                    self._num_files.value += 1
            rank += 1
        return None

    async def _scan_worker(self) -> None:
        """Drain _scan_queue and run _ingest_dir concurrently (bounded pool)."""
        while True:
            item = await self._scan_queue.get()  # blocks
            if item is None:  # sentinel -> exit
                # do not task_done() for sentinel
                break
            drs_type, path, iterable, is_versioned = item
            try:
                await self._ingest_dir(
                    drs_type, path, iterable=iterable, is_versioned=is_versioned
                )
            except Exception as error:
                logger.error(error)
            finally:
                self._scan_queue.task_done()

    async def _iter_content(
        self,
        drs_type: str,
        inp_dir: str,
        pos: int = 0,
        is_versioned: bool = True,
    ) -> None:
        """Walk recursively until files or the version level is reached."""
        if self._test_env():
            return
        store = self.config.datasets[drs_type].backend
        suffix = "." + inp_dir.rpartition(".")[-1]
        try:
            is_file = await store.is_file(inp_dir)
            iterable = await store.is_dir(inp_dir)
        except Exception as error:
            logger.error("Error checking file %s", error)
            return

        iterable = False if suffix == ".zarr" else iterable
        op: Optional[Callable[..., Coroutine[Any, Any, None]]] = None
        if is_file and suffix in self.config.suffixes:
            op = self._ingest_dir
        elif pos <= 0 or suffix == ".zarr":
            op = self._ingest_dir

        if op is not None:
            # enqueue the heavy scan; workers will run _ingest_dir concurrently
            await self._scan_queue.put(
                (drs_type, inp_dir, iterable, is_versioned)
            )
            return

        # otherwise, recurse sequentially (cheap) — no task per directory
        try:
            async for sub in store.iterdir(inp_dir):
                await self._iter_content(
                    drs_type, sub, pos - 1, is_versioned=is_versioned
                )
        except Exception as error:
            logger.error(error)

    async def ingest_data(self) -> None:
        """Produce scan tasks and process them with a bounded worker pool."""
        if self.ingest_queue.silent is False:
            self._print_status.set()
            self._num_files.value = 0
            print_performance(
                self._print_status,
                self._num_files,
                self.ingest_queue.queue,
                self.ingest_queue.num_objects,
            )

        async with asyncio.TaskGroup() as tg:
            # start scan workers
            for _ in range(self._scan_concurrency):
                tg.create_task(self._scan_worker())

            # produce scan items by walking roots sequentially
            for drs_type, path in self.search_objects:  # <- property is sync
                pos, is_versioned = self.config.max_directory_tree_level(
                    path, drs_type=drs_type
                )
                if pos < 0:
                    logger.warning(
                        "Can't define latest version of versioned dataset."
                        " This might lead to unexpected results. Try adjusting"
                        " your search path."
                    )

                await self._iter_content(
                    drs_type, path, pos, is_versioned=is_versioned
                )

            # wait until all queued scan items are processed
            await self._scan_queue.join()

            # stop workers (one sentinel per worker)
            for _ in range(self._scan_concurrency):
                await self._scan_queue.put(None)

        logger.info(
            "%i ingestion tasks have been completed", len(self._search_objects)
        )
        self.ingest_queue.join_all_tasks()
        if self.ingest_queue.silent is False:
            self._print_status.clear()



"""Interact with an S3 Object Store."""

import asyncio
import pathlib
from typing import AsyncIterator, Optional, Tuple, Union, cast

import fsspec
from s3fs import S3FileSystem

from ..api.storage_backend import MetadataType, PathTemplate
from ..logger import logger


class S3Path(PathTemplate):
    """Class to interact with an S3 object store."""

    _fs_type = "s3"

    def __post_init__(self) -> None:
        self._client: Optional[S3FileSystem] = None
        self.storage_options = self.storage_options or {"anon": True}

    async def close(self) -> None:
        """Close the connection."""
        client = await self._get_client()
        await client.s3.close()

    def get_fs_and_path(self, path: str) -> Tuple[fsspec.AbstractFileSystem, str]:
        """S3 implementation for returning (fs, path) suitable for xarray.

        Parameters
        ^^^^^^^^^^
        path:
            Path to the object store / file name

        Returns
        ^^^^^^^
        fsspec.AbstractFileSystem, str:
            The AbstractFileSystem class and the corresponding path to the
            data store.
        """
        return fsspec.filesystem("s3", **self.storage_options), path

    async def _get_client(self) -> S3FileSystem:
        if self._client is None:
            logger.debug(
                "Creating S3 Filesystem with storage_options: %s",
                self.storage_options,
            )
            loop = asyncio.get_running_loop()
            self._client = S3FileSystem(
                asynchronous=True, loop=loop, **self.storage_options
            )
            self._client._loop = loop

            await self._client.set_session()
        return self._client

    async def is_file(self, path: Union[str, pathlib.Path]) -> bool:
        """Check if a given path is a file object on the storage system."""
        client = await self._get_client()
        return cast(bool, await client._isfile(str(path)))

    async def is_dir(self, path: Union[str, pathlib.Path]) -> bool:
        """Check if a given path is a directory object on the storage system."""
        client = await self._get_client()
        return cast(bool, await client._isdir(str(path)))

    async def iterdir(self, path: Union[str, pathlib.Path]) -> AsyncIterator[str]:
        """Retrieve sub directories of directory."""
        client = await self._get_client()
        path = str(path)
        if await self.is_file(path):
            yield path
        else:
            for _content in await client._lsdir(path):
                size: int = _content.get("size") or 0
                if _content.get("type", "") == "directory" or size > 0:
                    yield _content.get("name", "")

    async def rglob(
        self, path: Union[str, pathlib.Path], glob_pattern: str = "*"
    ) -> AsyncIterator[MetadataType]:
        """Search recursively for files matching a ``glob_pattern``.

        Parameters
        ^^^^^^^^^^
        path: str
            A resource composed by:
                - bucket, 'bucketname'
                - prefix, 'prefix/to/a/path'
            E.g.: '/bucketname/prefix/to/objects'
            Will be translated into a request to
            `self.url`+`/bucketname?prefix="prefix/to/objects`
        glob_pattern: str
            A string reprenseting several glob patterns, separated by '|'
            E.g.: '*.zarr|*.nc|*.hdf5'
        """
        client = await self._get_client()
        if await self.is_file(path):
            yield MetadataType(path=str(path), metadata={})
        else:
            for suffix in self.suffixes:
                for content in await client._glob(f"{path}/**/*{suffix}"):
                    yield MetadataType(path=f"/{content}", metadata={})

    def path(self, path: Union[str, pathlib.Path]) -> str:
        """Get the full path (including any schemas/netlocs).

        Parameters
        ^^^^^^^^^^
        path: str, pathlib.Path
            Path of the object store

        Returns
        ^^^^^^^
        str:
            URI of the object store
        """
        return cast(
            str, fsspec.filesystem("s3", **self.storage_options).url(str(path))
        )

    def uri(self, path: Union[str, pathlib.Path]) -> str:
        """Get the uri of the object store.

        Parameters
        ^^^^^^^^^^
        path: str, pathlib.Path
            Path of the object store

        Returns
        ^^^^^^^
        str:
            URI of the object store
        """
        return self.path(path)



# tests/test_cli.py
from __future__ import annotations

import asyncio
from dataclasses import dataclass, field
from typing import (
    Annotated,
    Any,
    AsyncIterator,
    Callable,
    Dict,
    List,
    Optional,
    Tuple,
    cast,
)

import pytest

import metadata_crawler.cli as mc_cli

# -----------------------------
# Helpers / fakes for patching
# -----------------------------


@dataclass
class Call:
    fn: str
    args: Tuple[Any, ...] = field(default_factory=tuple)
    kwargs: Dict[str, Any] = field(default_factory=dict)


class Recorder:
    """Collects calls for assertions."""

    def __init__(self) -> None:
        self.calls: List[Call] = []

    def record(self, fn: str) -> Callable[..., Any]:
        def _inner(*args: Any, **kwargs: Any) -> Any:
            self.calls.append(Call(fn=fn, args=args, kwargs=kwargs))
            return None

        return _inner


class FakeConfig:
    def __init__(self, doc: Dict[str, Any]) -> None:
        self.merged_doc = doc

    def dumps(self) -> str:
        return "CONFIG_AS_TEXT"


class FakeIntakePath:
    def __init__(self, **storage_options: Any) -> None:
        self.storage_options = storage_options
        self.walk_called_with: Optional[str] = None

    async def walk(self, path: str) -> AsyncIterator[str]:
        # store the path so tests can assert
        self.walk_called_with = path
        for i in range(10):
            yield i


# Fake plugin class to exercise the dynamic subparsers
class DummyIngester:
    """dummy ingester"""

    # Annotated parameters -> the CLI builder reads "args" and other kwargs from the dict
    def index(
        self,
        path: Annotated[str, {"args": ["--path"], "help": "Path to data"}],
        limit: Annotated[int, {"args": ["--limit"], "help": "Limit"}] = 10,
    ) -> None:
        pass

    def delete(
        self,
        pattern: Annotated[str, {"args": ["--pattern"], "help": "Glob"}],
    ) -> None:
        pass


# Mark methods with _cli_help so your CLI includes them
DummyIngester.index._cli_help = "Index data"
DummyIngester.delete._cli_help = "Delete data"


# -----------------------------
# Tests
# -----------------------------


def test_crawl_parsing_and_dispatch(
    monkeypatch: pytest.MonkeyPatch, capsys: pytest.CaptureFixture[str]
) -> None:
    rec = Recorder()

    # Patch the functions used by the CLI
    monkeypatch.setattr(mc_cli, "add", rec.record("add"))
    monkeypatch.setattr(
        mc_cli, "get_config", lambda _cfg: FakeConfig({"ok": True})
    )
    monkeypatch.setattr(
        mc_cli, "load_plugins", lambda ep: {}
    )  # no plugin subcommands in this test
    # Prevent real file logging side effects

    # Build and parse args for the "add" subcommand
    parser = mc_cli.ArgParse()
    args = parser.parse_args(
        [
            "add",
            "s3://bucket/catalog.yml",
            "--batch-size",
            "42",
            "--data-set",
            "cmip6",
            "--data-set",
            "icon",
            "--data-object",
            "foo.yaml",
            "--n-procs",
            "4",
            "-s",
            "anon",
            "true",
            "-s",
            "retries",
            "3",
            "-s",
            "timeout",
            "1.5",
            "-s",
            "endpoint_url",
            "http://minio:9000",
            "--shadow",
            "foo",
            "bar",
            "--shadow",
            "1",
            "2",
            "-v",
            "-v",
        ]
    )

    # Execute the apply function
    mc_cli._run(args, **parser.kwargs)

    # Assert that `add` was called once with parsed kwargs
    assert len(rec.calls) == 1 and rec.calls[0].fn == "add"
    kw = rec.calls[0].kwargs

    # Core arguments
    assert kw["store"] == "s3://bucket/catalog.yml"
    assert kw["batch_size"] == 42
    assert kw["n_procs"] == 4
    assert kw["data_set"] == ["cmip6", "icon"]
    assert kw["data_object"] == ["foo.yaml"]

    # Storage options should be type-coerced
    so = kw["storage_options"]
    assert so == {
        "anon": True,
        "retries": 3,
        "timeout": 1.5,
        "endpoint_url": "http://minio:9000",
    }
    # The CLI injects verbosity into kwargs
    assert kw["verbosity"] == 2


def test_config_subcommand_prints_text(
    monkeypatch: pytest.MonkeyPatch, capsys: pytest.CaptureFixture[str]
) -> None:
    # Patch config + logging
    monkeypatch.setattr(mc_cli, "get_config", lambda _cfg: FakeConfig({"x": 1}))
    monkeypatch.setattr(mc_cli, "load_plugins", lambda ep: {})

    # Run full CLI entrypoint
    mc_cli.cli(["config", "-c", "conf.toml"])

    out = capsys.readouterr().out
    assert "CONFIG_AS_TEXT" in out  # display_config prints cfg.dumps() by default


def test_config_subcommand_prints_json(
    monkeypatch: pytest.MonkeyPatch, capsys: pytest.CaptureFixture[str]
) -> None:
    monkeypatch.setattr(mc_cli, "get_config", lambda _cfg: FakeConfig({"x": 1}))
    monkeypatch.setattr(mc_cli, "load_plugins", lambda ep: {})

    mc_cli.cli(["config", "-c", "conf.toml", "--json"])

    out = capsys.readouterr().out
    assert '"x": 1' in out


def test_walk_intake_invokes_async_walk(monkeypatch: pytest.MonkeyPatch) -> None:
    # Patch IntakePath and asyncio.run to avoid running a real loop
    fake_ip = FakeIntakePath()

    def fake_ctor(**storage_options: Any) -> FakeIntakePath:
        # keep the one instance so we can assert inside the fake run
        nonlocal fake_ip
        return fake_ip

    def fake_run(coro: Any) -> int:
        # Execute the coroutine on a fresh loop to avoid interference
        loop = asyncio.new_event_loop()
        try:
            loop.run_until_complete(coro)
        finally:
            loop.close()
        return 1

    monkeypatch.setattr(mc_cli, "IntakePath", fake_ctor)  # class replacement
    monkeypatch.setattr(mc_cli.asyncio, "run", fake_run)
    monkeypatch.setattr(mc_cli, "load_plugins", lambda ep: {})

    mc_cli.cli(["walk-intake", "s3://bucket/catalog.yaml", "-s", "anon", "true"])

    # Ensure our async method got called with the path
    assert fake_ip.walk_called_with == "s3://bucket/catalog.yaml"
    assert fake_ip.storage_options == {}


def test_plugin_index_and_delete_wiring(monkeypatch: pytest.MonkeyPatch) -> None:
    """
    Validates that:
      - load_plugins() drives dynamic subparsers
      - Annotated[...] metadata is converted to argparse options
      - apply_func is bound to top-level index/delete with correct kwargs
    """
    rec = Recorder()

    # Patch top-level functions that CLI dispatches to
    monkeypatch.setattr(
        mc_cli, "index", lambda *a, **kw: rec.calls.append(Call("index", a, kw))
    )
    monkeypatch.setattr(
        mc_cli, "delete", lambda *a, **kw: rec.calls.append(Call("delete", a, kw))
    )
    monkeypatch.setattr(
        mc_cli, "load_plugins", lambda ep: {"dummy": DummyIngester}
    )

    # Build CLI with plugin subcommands
    parser = mc_cli.ArgParse()

    # --- index path ---
    args = parser.parse_args(
        [
            "dummy",
            "index",
            "--path",
            "/data",
            "--limit",
            "5",
            "cat1.yml",
            "cat2.yml",
            "-s",
            "anon",
            "true",
        ]
    )
    mc_cli._run(args, **parser.kwargs)

    assert rec.calls and rec.calls[-1].fn == "index"
    kw = rec.calls[-1].kwargs
    # The CLI forwards parsed params using the parameter names (dest=param_name)
    assert kw["path"] == "/data"
    assert kw["limit"] == 5
    assert kw["catalogue_files"] == ["cat1.yml", "cat2.yml"]
    assert kw["index_system"] == "dummy"

    # --- delete path ---
    args = parser.parse_args(["dummy", "delete", "--pattern", "ua_*.nc"])
    mc_cli._run(args, **parser.kwargs)

    assert rec.calls and rec.calls[-1].fn == "delete"
    kw = rec.calls[-1].kwargs
    assert kw["pattern"] == "ua_*.nc"
    assert kw["index_system"] == "dummy"


def test_run_routes_exceptions_to_exception_handler(
    monkeypatch: pytest.MonkeyPatch,
) -> None:
    import metadata_crawler.utils as mc_utils

    class Boom(Exception):
        pass

    def raises(**_: Any) -> None:
        raise Boom("explode")

    ns = type("NS", (), {"apply_func": raises})  # argparse.Namespace-like

    # Fake logger to capture calls
    class FakeLogger:
        def __init__(self, level: int) -> None:
            self.level = level
            self.records: list[tuple[str, Any]] = []

        def error(self, msg: str, *, exc_info: Any = None) -> None:
            self.records.append((msg, exc_info))

        def critical(self, msg: str, *, exc_info: Any = None) -> None:
            self.records.append((msg, exc_info))

    # --- branch 1: level > 30 -> suffix added, exc_info=None ---
    high_logger = FakeLogger(level=50)
    monkeypatch.setattr(mc_utils, "logger", high_logger, raising=True)

    with pytest.raises(SystemExit) as ei1:
        mc_cli._run(cast(mc_cli.argparse.Namespace, ns), foo=1)

    assert ei1.value.code == 1
    assert len(high_logger.records) == 1
    msg1, exc1 = high_logger.records[0]
    assert "explode" in msg1
    assert msg1.endswith("increase verbosity for more information")
    assert exc1 is None

    # --- branch 2: level <= 30 -> no suffix, exc_info is exception ---
    low_logger = FakeLogger(level=10)
    monkeypatch.setattr(mc_utils, "logger", low_logger, raising=True)

    with pytest.raises(SystemExit) as ei2:
        mc_cli._run(cast(mc_cli.argparse.Namespace, ns), foo=1)

    assert ei2.value.code == 1
    assert len(low_logger.records) == 1
    msg2, exc2 = low_logger.records[0]
    assert msg2 == "explode"
    assert isinstance(exc2, Boom)


def test_process_storage_option_behavior() -> None:
    assert mc_cli._process_storage_option("true") is True
    assert mc_cli._process_storage_option("FALSE") is False
    assert mc_cli._process_storage_option("3") == 3
    assert mc_cli._process_storage_option("1.25") == 1.25
    assert mc_cli._process_storage_option("http://x") == "http://x"



"""API for adding new storage backends via :py:class:`BasePath`."""

import abc
import os
import pathlib
import threading
from getpass import getuser
from typing import (
    Any,
    AsyncIterator,
    ClassVar,
    Dict,
    List,
    Optional,
    TypedDict,
    Union,
    cast,
)

import h5netcdf
import xarray as xr
from pydantic import BaseModel, Field

from .mixin import LookupMixin, PathMixin, TemplateMixin


class MetadataType(TypedDict):
    """A dict representation of the metadata."""

    path: str
    metadata: Dict[str, Any]


class Metadata(BaseModel):
    """Meta data that is attached to each discovered path."""

    path: str
    metadata: Dict[str, Any] = Field(default_factory=dict)


class BasePath(abc.ABCMeta):
    """Every storage backend class should be of this type."""


class PathTemplate(
    abc.ABC, PathMixin, TemplateMixin, LookupMixin, metaclass=BasePath
):
    """Base class for interacting with different storage systems.

    This class defines fundamental methods that should be implemented
    to retrieve information across different storage systems.

    Parameters
    ^^^^^^^^^^
    suffixes: List[str], default:  [".nc", ".girb", ".zarr", ".tar", ".hdf5"]
        A list of available file suffixes.

    Other Parameters
    ^^^^^^^^^^^^^^^^
    storage_options: Any
        Information needed to interact with the storage system.

    Attributes
    ^^^^^^^^^^
    _user : str
        Value of the ``DRS_STORAGE_USER`` env variable (defaults to current user)
    _pw : str
        a password passed by the ``DRS_STORAGE_PASSWD`` env variable
    suffixes: List[str]
        A list of available file suffixes.
    storage_options: Dist[str, Any]
        A dict with information needed to interact with the storage system.
    """

    _fs_type: ClassVar[Optional[str]]
    """Definition of the file system time for each implementation."""

    _lock = threading.RLock()

    def __init__(
        self, suffixes: Optional[List[str]] = None, **storage_options: Any
    ) -> None:

        self._user: str = os.environ.get("DRS_STORAGE_USER") or getuser()
        self._pw: str = os.environ.get("DRS_STORAGE_PASSWD") or ""
        self.suffixes = suffixes or [".nc", ".girb", ".zarr", ".tar", ".hdf5"]
        self.storage_options = cast(
            Dict[str, Any], self.render_templates(storage_options or {}, {})
        )
        self.set_static_from_nested()
        self.__post_init__()

    def __post_init__(self) -> None:
        """Call this method after the __init__ get called.

        If you need to assign any attributes redefine this method in your class.
        """

    async def close(self) -> None:
        """Close any open sessions."""

    def open_dataset(
        self, path: str, **read_kws: Any
    ) -> Union[xr.Dataset, h5netcdf.core.File]:
        """Open a dataset with xarray.

        Parameters
        ^^^^^^^^^^
        path:
            Path to the object store / file name
        **read_kws:
            Keyword arguments passed to open the datasets.

        Returns
        ^^^^^^-
        xarray.Dataset:
            The xarray dataset.
        """
        fs, path = self.get_fs_and_path(path)

        def _get_engine(file_name: str) -> str:
            engines = {
                "grb": "cfgrib",
                "grib": "cfgrib",
                "gb": "gb",
                "nc": "h5netcdf",
                "nc4": "h5netcdf",
                "netcdf": "h5netcdf",
                "cdf": "h5netcdf",
                "hdf5": "h5netcdf",
                "h5": "h5netcdf",
                "zarr": "zarr",
                "zar": "zarr",
            }
            suffix = file_name.rpartition(".")[-1]
            return engines.get(suffix, "")

        kwargs = read_kws.copy()
        engine = kwargs.setdefault("engine", _get_engine(path) or None)

        if engine == "zarr":
            dset: xr.Dataset = xr.open_zarr(fs.get_mapper(path))
            return dset
        if fs.protocol[0] == "file" and engine == "h5netcdf":
            return h5netcdf.File(path)
        if fs.protocol[0] == "file":
            return xr.open_mfdataset(path, **kwargs)
        with fs.open(path, "rb") as stream:
            return xr.open_dataset(stream, **kwargs)

    def read_attr(
        self, attribute: str, path: Union[str, pathlib.Path], **read_kws: Any
    ) -> Any:
        """Get a metadata attribute from a datastore object.

        Parameters
        ^^^^^^^^^^
        attr: The attribute that is queried can be of the form of
              <attribute>, <variable>.<attribute>, <attribute>,
              <variable>.<attribute>
        path: Path to the object store / file path
        read_kws: Keyword arguments for opening the datasets.

        Returns
        ^^^^^^^
        str: Metadata from the data.
        """
        with self.open_dataset(str(path), **read_kws) as dset:
            if "." not in attribute:
                return dset.attrs[attribute]
            var, _, attr = attribute.partition(".")
            return dset[var].attrs[attr]

    @abc.abstractmethod
    async def is_dir(self, path: Union[str, pathlib.Path]) -> bool:
        """Check if a given path is a directory object on the storage system.

        Parameters
        ^^^^^^^^^^
        path : str, pathlib.Path
            Path of the object store

        Returns
        ^^^^^^-
        bool: True if path is dir object, False if otherwise or doesn't exist
        """

    @abc.abstractmethod
    async def is_file(self, path: Union[str, pathlib.Path]) -> bool:
        """Check if a given path is a file object on the storage system.

        Parameters
        ^^^^^^^^^^
        path:
            Path of the object store

        Returns
        ^^^^^^^
        bool:
            True if path is file object, False if otherwise or doesn't exist
        """
        ...  # pragma: no cover

    @abc.abstractmethod
    async def iterdir(
        self,
        path: Union[str, pathlib.Path],
    ) -> AsyncIterator[str]:
        """Get all sub directories from a given path.

        Parameters
        ^^^^^^^^^^
        path:
            Path of the object store

        Yields
        ^^^^^^
        str:
            1st level sub directory
        """
        yield ""  # pragma: no cover

    @abc.abstractmethod
    async def rglob(
        self, path: Union[str, pathlib.Path], glob_pattern: str = "*"
    ) -> AsyncIterator[MetadataType]:
        """Search recursively for paths matching a given glob pattern.

        Parameters
        ^^^^^^^^^^
        path:
            Path of the object store
        glob_pattern: str
            Pattern that the target files must match

        Yields
        ^^^^^^
        MetadataType: Path of the object store that matches the glob pattern.
        """
        yield MetadataType(path="", metadata={})  # pragma: no cover

    def fs_type(self, path: Union[str, pathlib.Path]) -> str:
        """Define the file system type."""
        return self._fs_type or ""

    @abc.abstractmethod
    def path(self, path: Union[str, pathlib.Path]) -> str:
        """Get the full path (including any schemas/netlocs).

        Parameters
        ^^^^^^^^^^
        path:
            Path of the object store

        Returns
        ^^^^^^^
        str:
            URI of the object store
        """
        ...  # pragma: no cover

    @abc.abstractmethod
    def uri(self, path: Union[str, pathlib.Path]) -> str:
        """Get the uri of the object store.

        Parameters
        ^^^^^^^^^^
        path:
            Path of the object store

        Returns
        ^^^^^^^
        str:
            URI of the object store
        """
        ...  # pragma: no cover



"""Test crawling s3 stores."""

from pathlib import Path
from typing import Dict

import intake

from metadata_crawler import add


def test_crawl_s3_obs(
    drs_config_path: Path,
    storage_options: Dict[str, str],
) -> None:
    """Test crawling s3."""
    cat_file = "s3://test/metadata_crawler/tests/data.yml"
    add(
        drs_config_path,
        store=cat_file,
        data_store_prefix="s3://test/metadata_crawler/tests/metadata",
        batch_size=3,
        n_procs=1,
        data_set=["obs-s3"],
        storage_options=storage_options,
    )
    cat = intake.open_catalog(cat_file, storage_options=storage_options)
    assert len(cat.latest.read()) > 0


def test_crawl_s3_dir(
    drs_config_path: Path,
    storage_options: Dict[str, str],
) -> None:
    """Test crawling a flat directory."""
    cat_file = "s3://test/metadata_crawler/tests/data-flat.yml"
    inp_dir = (
        "s3://test/data/obs/observations/grid/CPC/CPC/cmorph/"
        "30min/atmos/30min/r1i1p1/v20210618/pr"
    )
    add(
        drs_config_path,
        store=cat_file,
        data_store_prefix="s3://test/metadata_crawler/tests/metadata",
        batch_size=3,
        n_procs=1,
        data_object=[inp_dir],
        storage_options=storage_options,
    )
    cat = intake.open_catalog(cat_file, storage_options=storage_options)
    assert len(cat.latest.read()) > 0


def test_crawl_s3_single_file(
    drs_config_path: Path,
    storage_options: Dict[str, str],
) -> None:
    """Test crawling a flat directory."""
    cat_file = "s3://test/metadata_crawler/tests/data-flat.yml"
    inp_file = (
        "s3://test/data/obs/observations/grid/CPC/CPC/cmorph"
        "/30min/atmos/30min/r1i1p1/v20210618/pr"
        "/pr_30min_CPC_cmorph_r1i1p1_201609020000-201609020030.nc"
    )
    add(
        drs_config_path,
        store=cat_file,
        data_store_prefix="s3://test/metadata_crawler/tests/metadata",
        batch_size=3,
        n_procs=1,
        data_object=[inp_file],
        storage_options=storage_options,
    )
    cat = intake.open_catalog(cat_file, storage_options=storage_options)
    assert len(cat.latest.read()) > 0


def test_crawl_s3_cmip6(
    drs_config_path: Path, storage_options: Dict[str, str]
) -> None:
    cat_file = "s3://test/metadata_crawler/tests/cmip6-s3.yml"

    add(
        drs_config_path,
        store=cat_file,
        data_store_prefix="s3://test/metadata_crawler/tests/cmip6-s3",
        batch_size=3,
        n_procs=1,
        data_set=["cmip6-s3"],
        verbosity=5,
        storage_options=storage_options,
    )
    cat = intake.open_catalog(cat_file, storage_options=storage_options)
    assert len(cat.latest.read()) > 0
    # There are versioned datasets so latest should not have all the entries
    assert len(cat.latest.read()) < len(cat.files.read())


def test_crawl_single_s3_file(
    drs_config_path: Path,
    cat_file: Path,
) -> None:
    """Test if we can crawl a single file on s3."""
    file = (
        "s3://test/data/model/global/cmip6/CMIP6/CMIP/CSIRO-ARCCSS/"
        "ACCESS-CM2/amip/r1i1p1f1/Amon/ua/gn/v20191108/"
        "ua_Amon_ACCESS-CM2_amip_r1i1p1f1_gn_197001-201512.nc"
    )
    add(
        drs_config_path,
        n_procs=1,
        store=cat_file,
        batch_size=3,
        data_object=[file],
    )
    assert cat_file.is_file()
    cat = intake.open_catalog(cat_file)
    assert len(cat.latest.read()) == len(cat.files.read()) == 1


def test_crawl_s3_zarr(drs_config_path: Path, cat_file: Path) -> None:
    """Zarr stores are directories but are treated as files."""

    add(
        drs_config_path,
        data_set=["nextgems-s3"],
        store=cat_file,
        n_procs=1,
        batch_size=3,
    )
    assert cat_file.is_file()
    cat = intake.open_catalog(cat_file)
    assert len(cat.latest.read()) > 0



"""Test crawling metadata from posix fs."""

import os
from pathlib import Path

import intake
import numpy as np
import pytest

from metadata_crawler import add
from metadata_crawler.api.config import DRSConfig
from metadata_crawler.utils import EmptyCrawl, MetadataCrawlerException


def test_crawl_local_obs(
    drs_config_path: Path, cat_file: Path, data_dir: Path
) -> None:
    """Test crawling the local observations."""
    _lens = []
    add(
        drs_config_path,
        store=cat_file,
        n_procs=1,
        batch_size=3,
        catalogue_backend="jsonlines",
        data_object=[data_dir / "observations"],
    )
    assert cat_file.exists()
    _lens.append(len(intake.open_catalog(cat_file).latest.read()))
    add(
        drs_config_path,
        store=cat_file,
        n_procs=1,
        batch_size=3,
        data_set=["obs-fs"],
        catalogue_backend="jsonlines",
    )
    assert cat_file.exists()
    _lens.append(len(intake.open_catalog(cat_file).latest.read()))
    assert _lens[0] == _lens[-1]
    with pytest.raises(MetadataCrawlerException):
        add(cat_file)


def test_crawl_cordex(drs_config_path: Path, cat_file: Path) -> None:
    """Test for domain specs in cordex."""
    add(
        drs_config_path.parent,
        store=cat_file,
        data_set="cordex-fs",
    )
    assert cat_file.exists()
    cfg = DRSConfig.load(drs_config_path)
    cat = intake.open_catalog(cat_file).latest.read()[0]
    assert "product" in cat
    assert "EUR-11" in cat["product"]
    assert "bbox" in cat
    np.testing.assert_allclose(
        np.array(cat["bbox"]), np.array(cfg.dialect["cordex"].domains["EUR-11"])
    )


def test_crawl_local_cmip6(
    drs_config_path: Path, cat_file: Path, data_dir: Path
) -> None:
    """Test carawling local cmip6."""
    _data_dir = data_dir / "model" / "global" / "cmip6"
    _lens = []
    add(
        drs_config_path,
        store=cat_file,
        n_procs=10,
        batch_size=20_000,
        data_object=[_data_dir],
    )
    assert cat_file.exists()
    _lens.append(len(intake.open_catalog(cat_file).latest.read()))
    add(
        drs_config_path,
        store=cat_file,
        n_procs=1,
        batch_size=3,
        data_set=["cmip6-fs"],
    )
    assert cat_file.exists()
    _lens.append(len(intake.open_catalog(cat_file).latest.read()))
    assert _lens[0] == _lens[-1]
    with pytest.raises(MetadataCrawlerException):
        add(cat_file, drs_config_path, data_set=["zzz"])


def test_crawl_empty_set(drs_config_path: Path, cat_file: Path) -> None:
    """Test the behaviour of crawling non existing data(sets)."""

    with pytest.raises(MetadataCrawlerException):
        add("foo.toml", store="foo.yaml")
    with pytest.raises(MetadataCrawlerException):
        add(drs_config_path, data_set=["zzz"], store="foo.yaml")
    with pytest.raises(MetadataCrawlerException):
        add(drs_config_path, store="foo.yaml", data_set=["nextgems_zarr"])
    with pytest.raises(MetadataCrawlerException):
        add(drs_config_path, store="foo.yaml", data_object=["/foo"])
    with pytest.raises(EmptyCrawl):
        add(drs_config_path, data_set=["fool"], store="foo.yam")


def test_crawl_single_files(
    drs_config_path: Path, cat_file: Path, data_dir: Path
) -> None:
    """Test if we can ingest a single file."""
    add(
        drs_config_path,
        store=cat_file,
        data_store_prefix=str(drs_config_path.parent / "foo"),
        verbosity=5,
        data_object=[
            os.path.join(
                data_dir,
                "intake",
                "work",
                "bm1235",
                "k202181",
                "ngc4008a",
                "ngc4008a_P1D_1.zarr",
            )
        ],
    )
    cat = intake.open_catalog(cat_file)
    assert len(cat.latest.read()) > 0
    _file = os.path.join(
        data_dir,
        "intake",
        "work",
        "bm1235",
        "a270046",
        "cycle3",
        "tco2559l137",
        "hzfy",
        "hres",
        "intel.levante.openmpi",
        "lvt.intel.sp",
        "Cycle3_012020",
        "atmoce_y.fesom.2020_cropped_restricted_compressed.nc",
    )
    add(
        drs_config_path,
        store=cat_file,
        data_object=[_file],
        verbosity=5,
    )
    cat = intake.open_catalog(cat_file)
    assert len(cat.latest.read()) > 0


def test_crawl_thresh_fail(drs_config_path: Path, cat_file: Path) -> None:
    """Test if we can't crawl under a certain threshold."""
    with pytest.raises(EmptyCrawl):
        add(
            drs_config_path,
            store=cat_file,
            data_set=["obs-fs-missing"],
            catalogue_backend="jsonlines",
            fail_under=10,
        )
    assert not cat_file.exists()



"""Logging utilities."""

import logging
import logging.config
import os
from logging.handlers import RotatingFileHandler
from pathlib import Path
from typing import Any, Optional, cast

import appdirs
from rich.console import Console
from rich.logging import RichHandler

THIS_NAME = "metadata-crawler"

logging.basicConfig(
    level=logging.WARNING,
    format="%(asctime)s %(levelname)s: %(name)s - %(message)s",
)

logging.config.dictConfig(
    {
        "version": 1,
        # keep existing handlers
        "disable_existing_loggers": False,
        "root": {
            "level": "CRITICAL",
            "handlers": ["default"],
        },
        "formatters": {
            "standard": {
                "format": "%(asctime)s %(levelname)s: %(name)s - %(message)s",
            },
        },
        "handlers": {
            "default": {
                "class": "logging.StreamHandler",
                "formatter": "standard",
                "level": "CRITICAL",
            },
        },
    }
)


class Logger(logging.Logger):
    """Custom Logger defining the logging behaviour."""

    logfmt: str = "%(name)s: %(message)s"
    filelogfmt: str = "%(asctime)s %(levelname)s: %(name)s - %(message)s"
    datefmt: str = "%Y-%m-%dT%H:%M:%S"
    no_debug: list[str] = ["watchfiles", "httpcore", "pymongo", "pika"]

    def __init__(
        self,
        name: Optional[str] = None,
        level: Optional[int] = None,
        suffix: Optional[str] = None,
    ) -> None:
        """Instantiate this logger only once and for all."""
        self.level = level or int(
            cast(str, os.getenv("MDC_LOG_LEVEL", str(logging.CRITICAL)))
        )
        name = name or THIS_NAME
        logger_format = logging.Formatter(self.logfmt, self.datefmt)
        self.file_format = logging.Formatter(self.filelogfmt, self.datefmt)
        self._logger_file_handle: Optional[RotatingFileHandler] = None
        self._logger_stream_handle = RichHandler(
            rich_tracebacks=True,
            tracebacks_max_frames=10,
            tracebacks_extra_lines=5,
            show_path=True,
            console=Console(
                soft_wrap=False,
                force_jupyter=False,
                stderr=True,
            ),
        )
        self._logger_stream_handle.setFormatter(logger_format)
        self._logger_stream_handle.setLevel(self.level)
        super().__init__(name, self.level)

        self.propagate = False
        self.handlers = [self._logger_stream_handle]
        (
            self.add_file_handle(suffix=suffix)
            if os.getenv("MDC_LOG_INIT", "0") == "1"
            else None
        )

    def set_level(self, level: int) -> None:
        """Set the logger level to level."""
        for handler in self.handlers:
            log_level = level
            if isinstance(handler, RotatingFileHandler):
                log_level = min(level, logging.CRITICAL)
            handler.setLevel(log_level)
        self.setLevel(level)
        self.level = level

    def error(
        self,
        msg: object,
        *args: Any,
        **kwargs: Any,
    ) -> None:
        """Log an error. When log level is smaller than INFO, log exceptions."""
        if self.level < logging.INFO:
            kwargs.setdefault("exc_info", True)
        self._log(logging.ERROR, msg, args, **kwargs)

    def add_file_handle(
        self,
        suffix: Optional[str] = None,
        level: int = logging.CRITICAL,
    ) -> None:
        """Add a file log handle to the logger."""
        suffix = suffix or os.getenv("MDC_LOG_SUFFIX", "")
        base_name = f"{THIS_NAME}-{suffix}" if suffix else THIS_NAME
        log_dir = Path(os.getenv("MDC_LOG_DIR", appdirs.user_log_dir(THIS_NAME)))
        log_dir.mkdir(exist_ok=True, parents=True)
        logger_file_handle = RotatingFileHandler(
            log_dir / f"{base_name}.log",
            mode="a",
            maxBytes=5 * 1024**2,
            backupCount=5,
            encoding="utf-8",
            delay=False,
        )
        logger_file_handle.setFormatter(self.file_format)
        logger_file_handle.setLevel(self.level)
        self.addHandler(logger_file_handle)


logger = Logger()


def get_level_from_verbosity(verbosity: int) -> int:
    """Calculate the log level from a verbosity."""
    return max(logging.CRITICAL - 10 * verbosity, -1)


def apply_verbosity(
    level: Optional[int] = None, suffix: Optional[str] = None
) -> int:
    """Set the logging level of the handlers to a certain level."""
    level = logger.level if level is None else level
    old_level = logger.level
    level = get_level_from_verbosity(level)
    logger.set_level(level)
    logger.add_file_handle(suffix, level)

    return old_level



"""test ingesting."""

import os
from pathlib import Path
from tempfile import TemporaryDirectory
from typing import Any, Dict, List

import intake
import mock
import requests
from pymongo import MongoClient

from metadata_crawler import add, delete, index


def test_ingest_solr(
    drs_config_path: Path, storage_options: Dict[str, str], solr_server: str
) -> None:
    """Test ingesting metadata to solr."""

    cat_file = "s3://test/metadata_crawler/tests/solr.yml"
    cat_files = []
    lens = []
    for ds in ("fs", "s3", "swift"):
        cat_file = f"s3://test/metadata_crawler/tests/solr-{ds}.yml"
        cat_files.append(cat_file)
        add(
            drs_config_path,
            store=cat_file,
            data_store_prefix=f"s3://test/metadata_crawler/tests/solrdata-{ds}",
            data_set=[f"obs-{ds}"],
            storage_options=storage_options,
        )
        cat = intake.open_catalog(cat_file, storage_options=storage_options)
        lens.append(len(cat.latest.read()))
    with mock.patch.dict(os.environ, {"MDC_INTERACTIVE": "1"}, clear=True):
        index(
            "solr",
            *cat_files,
            server=solr_server,
            storage_options=storage_options,
            batch_size=1,
        )
    res = requests.get(
        f"{solr_server}/solr/latest/select", params={"q": "*:*", "rows": 0}
    )
    num = res.json().get("response", {}).get("numFound", 0)
    assert sum(lens) == num


def test_ingest_mongo(
    drs_config_path: Path,
    mongo_server: Dict[str, str],
) -> None:
    """Test ingesting metadata to mongo."""

    cat_files = []
    lens = []
    with TemporaryDirectory() as temp_dir:
        for ds in ("fs", "s3", "swift"):
            cat_file = Path(temp_dir) / f"mongo-{ds}.yml"
            cat_files.append(cat_file)
            add(
                drs_config_path,
                store=cat_file,
                data_store_prefix=f"{ds}-mongodata",
                data_set=[f"obs-{ds}"],
                catalogue_backend="jsonlines",
            )
            cat = intake.open_catalog(cat_file)
            lens.append(len(cat.latest.read()))
        index("mongo", *cat_files, **mongo_server)
    with MongoClient(mongo_server["url"]) as client:
        col = client[mongo_server["database"]]["latest"]
        _f = list(col.find({}))
        print(len(_f))
        assert len(_f) == sum(lens)


def test_delete_mongo(
    mongo_server: Dict[str, str],
    metadata: List[Dict[str, Any]],
) -> None:
    """Test deleting metadata from mongo."""

    with MongoClient(mongo_server["url"]) as client:
        db = client[mongo_server["database"]]
        for col in ("files", "latest"):
            collection = db[col]
            for md in metadata:
                collection.insert_one(md)
    delete("mongo", facets=[("project", "*")], **mongo_server)
    with MongoClient(mongo_server["url"]) as client:
        col = client[mongo_server["database"]]["latest"]
        assert len(list(col.find({}))) == 0
    delete("mongo", facets=[("project", "foo")], **mongo_server)
    delete("mongo", **mongo_server)


def test_delete_solr(
    solr_server: str,
    metadata: List[Dict[str, Any]],
) -> None:
    """Test deleting metadata from solr."""
    for core in ("latest", "files"):
        url = f"{solr_server}/solr/{core}/update/json?commit=true"
        res = requests.post(url, json=metadata)
        res.raise_for_status()
    delete("solr", facets=[("project", "*")], server=solr_server)
    res = requests.get(
        f"{solr_server}/solr/latest/select", params={"q": "*:*", "rows": 0}
    )
    num = res.json().get("response", {}).get("numFound", 0)
    assert num == 0
    delete("solr", facets=[("file", "/foo/*")], server=solr_server)
    delete("solr", facets=[("file", "/foo/*")], server=solr_server)



"""API for loading crawler configuration."""

from __future__ import annotations

import glob
import os
import re
import textwrap
from copy import deepcopy
from datetime import datetime
from enum import Enum, StrEnum
from fnmatch import fnmatch
from pathlib import Path
from typing import (
    Annotated,
    Any,
    Dict,
    List,
    Literal,
    Optional,
    Tuple,
    Union,
    cast,
)
from urllib.parse import urlsplit
from warnings import catch_warnings

import tomli
import tomlkit
import xarray
from pydantic import (
    BaseModel,
    ConfigDict,
    Field,
    ValidationError,
    field_validator,
    model_validator,
)
from tomlkit.container import OutOfOrderTableProxy
from tomlkit.items import Table

from ..utils import (
    MetadataCrawlerException,
    convert_str_to_timestamp,
    load_plugins,
)
from ..utils.cftime_utils import infer_cmor_like_time_frequency
from .mixin import TemplateMixin
from .storage_backend import Metadata, MetadataType


class BaseType(str, Enum):
    """Basic types."""

    string = "string"
    integer = "integer"
    float = "float"
    timestamp = "timestamp"


class Types(str, Enum):
    """Types supported by the config."""

    string = "string"
    integer = "integer"
    float = "float"
    timestamp = "timestamp"
    daterange = "timestamp[2]"
    path = "string"
    uri = "string"
    dataset = "string"
    fmt = "string"
    storage = "string"
    bbox = "float[4]"


class SchemaField(BaseModel):
    """BaseModel defining the metadata schema."""

    key: str
    type: str
    required: bool = False
    default: Optional[Any] = None
    length: Optional[int] = None
    base_type: BaseType = BaseType.string
    multi_valued: bool = True
    indexed: bool = True
    name: Optional[str] = None
    unique: bool = False

    @field_validator("type")
    @classmethod
    def parse_type(cls, v: str) -> str:
        """Parse the data types.

        Accepts
        ^^^^^^^
          - 'string', 'integer', 'float', 'timestamp' -> length=None
          - 'float[2]', 'int[5]' -> length=number
          - 'string[]'  -> length=None, multi_valued semantics
          - 'daterange' [timestamp, timestamp]
        """
        f_type = getattr(getattr(Types, v, None), "value", v)
        m = re.fullmatch(r"({})(\[(\d*)\])?".format("|".join(BaseType)), f_type)
        if not m:
            raise MetadataCrawlerException(f"invalid type spec {v!r}")
        base, _, num = m.groups()
        setattr(
            cls,
            "__parsed_type",
            {"base": base, "length": int(num) if num else None},
        )
        return v

    @model_validator(mode="after")
    def _set_parsed(self) -> "SchemaField":
        parsed = getattr(self, "__parsed_type")
        self.base_type = BaseType(parsed["base"])
        self.length = parsed["length"]
        self.name = self.name or self.key
        return self

    @staticmethod
    def get_time_range(
        time_stamp: Optional[Union[str, List[str]]],
    ) -> List[datetime]:
        """Convert a from to time range to a begin or end time step."""
        time_stamp = time_stamp or ""
        if isinstance(time_stamp, str):
            start_str, _, end_str = (
                time_stamp.replace(":", "").replace("_", "-").partition("-")
            )
            time_stamp = [start_str or "fx", end_str or "fx"]
        for n, ts in enumerate(time_stamp):
            if hasattr(ts, "isoformat"):
                time_stamp[n] = ts.isoformat()
            time_stamp[n] = str(ts) or "fx"
        start = convert_str_to_timestamp(
            time_stamp[0], alternative="0001-01-01T00:00"
        )
        end = convert_str_to_timestamp(
            time_stamp[-1], alternative="9999-12-31T23:59"
        )
        return [start, end]


class MetadataSource(StrEnum):
    """Representation of how the metadata should be retrieved."""

    storage = "storage"  # via intake/fdb5/etc.
    path = "path"  # parse via specs_dir/specs_file
    data = "data"  # read attributes from the file itself


class VarAttrRule(BaseModel):
    """How to read attributes from variables."""

    var: str
    attr: str  # attribute name on DataArray.attrs
    default: Any = None


class StatRule(BaseModel):
    """How to apply statistics."""

    stat: Literal["min", "max", "minmax", "range", "bbox", "timedelta"]
    var: Optional[str] = None  # for numeric stats on a single var
    coords: Optional[Union[List[str], str]] = None  # for time range, etc.
    lat: Optional[str] = None  # convenience keys for bbox
    lon: Optional[str] = None
    default: Any = None


class ConfigMerger:
    """Load the system and user TOML, merges user -> system under.

    Merging the config preserves comments/formatting, and lets you inspect or
    write the result.
    """

    def __init__(
        self,
        *user_paths_or_config: Union[
            Path, str, Dict[str, Any], tomlkit.TOMLDocument
        ],
    ):
        # parse both documents
        system_path = Path(__file__).parent / "drs_config.toml"
        self._system_doc = tomlkit.parse(system_path.read_text(encoding="utf-8"))
        _configs: List[str] = []
        for user_path_or_config in user_paths_or_config:
            if isinstance(user_path_or_config, (str, Path)) and os.path.isdir(
                user_path_or_config
            ):
                _configs.append(
                    (
                        Path(user_path_or_config).expanduser().absolute()
                        / "drs_config.toml"
                    ).read_text(encoding="utf-8")
                )
            elif isinstance(user_path_or_config, (str, Path)) and os.path.isfile(
                user_path_or_config
            ):
                _configs.append(
                    (
                        Path(user_path_or_config)
                        .expanduser()
                        .absolute()
                        .read_text(encoding="utf-8")
                    )
                )
            elif isinstance(user_path_or_config, (str, Path)):
                paths_or_cfg = glob.glob(
                    str(user_path_or_config), recursive=False
                ) or [str(user_path_or_config)]
                for path_or_cfg in paths_or_cfg:
                    if os.path.isfile(path_or_cfg):
                        _configs.append(
                            Path(path_or_cfg).read_text(encoding="utf-8")
                        )
                    # We have most likely a string representing a config.
                    elif not os.path.exists(path_or_cfg):
                        _configs.append(str(user_path_or_config))
            else:
                _configs.append(tomlkit.dumps(user_path_or_config))
        for _config in _configs:
            try:
                self._user_doc = tomlkit.parse(_config)
            except Exception:
                raise MetadataCrawlerException("Could not load config path.")
            self._merge_tables(self._system_doc, self._user_doc)

    def _merge_tables(
        self,
        base: Union[tomlkit.TOMLDocument, Table, OutOfOrderTableProxy],
        override: Union[Table, tomlkit.TOMLDocument, OutOfOrderTableProxy],
    ) -> None:

        for key, value in override.items():
            if key not in base:
                base[key] = value
                continue
            if isinstance(value, (Table, OutOfOrderTableProxy)):
                self._merge_tables(cast(Table, base[key]), value)
            else:
                base[key] = value

    @property
    def merged_doc(self) -> tomlkit.TOMLDocument:
        """Return the merged TOMLDocument."""
        return self._system_doc

    def dumps(self) -> str:
        """Return the merged document as a TOML string."""
        return tomlkit.dumps(self.merged_doc)


def strip_protocol(inp: str | Path) -> Path:
    """Extract the path from a given input file system."""
    abs_path = Path(urlsplit(str(inp)).path).expanduser()
    return Path(abs_path)


class CrawlerSettings(BaseModel):
    """Define the user input for a data crawler session."""

    name: str
    search_path: Union[str, Path]

    def model_post_init(self, __context: Any = None) -> None:
        """Apply rules after init."""
        self.search_path = str(self.search_path)


class PathSpecs(BaseModel):
    """Implementation of the Directory reference syntax."""

    dir_parts: Optional[List[str]] = None
    file_parts: Optional[List[str]] = None
    file_sep: str = "_"

    def _get_metadata_from_dir(
        self, data: Dict[str, Any], rel_path: Path
    ) -> None:
        dir_parts = rel_path.parent.parts

        if self.dir_parts and len(dir_parts) == len(self.dir_parts):
            data.update(
                {
                    k: v
                    for (k, v) in zip(self.dir_parts, dir_parts)
                    if k not in data
                }
            )
        elif self.dir_parts:
            raise MetadataCrawlerException(
                (
                    f"Number of dir parts for {rel_path.parent} do not match "
                    f"- needs: {len(self.dir_parts)} has: {len(dir_parts)}"
                )
            ) from None

    def _get_metadata_from_filename(
        self, data: Dict[str, Any], rel_path: Path
    ) -> None:
        if self.file_parts is None:
            return
        file_parts = rel_path.name.split(self.file_sep)
        _parts: Dict[str, str] = {}
        if len(file_parts) == len(self.file_parts):
            _parts = dict(zip(self.file_parts, file_parts))
        elif (
            len(file_parts) == len(self.file_parts) - 1 and "fx" in rel_path.name
        ):
            _parts = dict(zip(self.file_parts[:-1], file_parts))
        else:
            raise MetadataCrawlerException(
                (
                    f"Number of file parts for {rel_path.name} do not match "
                    f"- needs: {len(self.file_parts)} has: {len(file_parts)})"
                )
            )
        data.update({k: v for (k, v) in _parts.items() if k not in data})

    def get_metadata_from_path(self, rel_path: Path) -> Dict[str, Any]:
        """Read path encoded metadata from path specs."""
        data: Dict[str, Any] = {}
        self._get_metadata_from_dir(data, rel_path)
        self._get_metadata_from_filename(data, rel_path)
        data.pop("_", None)
        return data


class DataSpecs(BaseModel):
    """BaseModel for the configuration."""

    globals: Dict[str, str] = Field(default_factory=dict)
    var_attrs: Dict[str, VarAttrRule] = Field(default_factory=dict)
    stats: Dict[str, StatRule] = Field(default_factory=dict)
    read_kws: Dict[str, Any] = Field(default_factory=dict)

    def _set_global_attributes(
        self, dset: "xarray.Dataset", out: Dict[str, Any]
    ) -> None:

        for facet, attr in self.globals.items():
            if attr == "__variable__":
                out[facet] = list(getattr(dset, "data_vara", dset.variables))
            else:
                out[facet] = dset.attrs.get(attr)

    def _set_variable_attributes(
        self, dset: "xarray.Dataset", out: Dict[str, Any]
    ) -> None:
        data_vars = list(getattr(dset, "data_vars", dset.variables))

        def get_val(
            rule: VarAttrRule, vnames: Union[str, List[str]]
        ) -> List[Any]:
            if isinstance(vnames, str):
                vnames = [dv for dv in data_vars if fnmatch(dv, vnames)]
            attr_list: List[Any] = []
            for vname in vnames:
                default = (rule.default or "").replace("__name__", vname) or vname
                attr = (rule.attr or "").replace("__name__", vname) or vname
                if vname in dset:
                    attr_list.append(dset[vname].attrs.get(attr, default))
                else:
                    attr_list.append(default)
            return attr_list

        for facet, rule in self.var_attrs.items():
            resolved: Union[str, List[str]] = rule.var
            vals = get_val(rule, resolved)
            if len(vals) == 1:
                out[facet] = vals[0]
            else:
                out[facet] = vals

    def _apply_stats_rules(
        self, dset: "xarray.Dataset", out: Dict[str, Any]
    ) -> None:

        for facet, rule in self.stats.items():
            coords: Optional[List[str]] = None
            if rule.coords:
                coords = (
                    rule.coords
                    if isinstance(rule.coords, list)
                    else [rule.coords]
                )
            match rule.stat:
                case "bbox":
                    lat = rule.lat or (coords[0] if coords else "lat")
                    lon = rule.lon or (
                        coords[1] if coords and len(coords) > 1 else "lon"
                    )
                    out[facet] = rule.default
                    if lat in dset and lon in dset:
                        latv = dset[lat].values
                        lonv = dset[lon].values
                        out[facet] = [
                            float(lonv.min()),
                            float(lonv.max()),
                            float(latv.min()),
                            float(latv.max()),
                        ]

                case "range":
                    coord = coords[0] if coords else None
                    out[facet] = rule.default
                    if coord and coord in dset.coords:
                        arr = dset.coords[coord].values
                        out[facet] = [arr.min(), arr.max()]

                case "min" | "max" | "minmax":

                    coord = coords[0] if coords else None
                    var_name = rule.var if rule.var else coord
                    out[facet] = rule.default
                    if var_name and var_name in dset:
                        arr = dset[var_name].values
                        if rule.stat == "min":
                            out[facet] = arr.min()
                        elif rule.stat == "max":
                            out[facet] = arr.max()
                        else:
                            out[facet] = [arr.min(), arr.max()]
                case "timedelta":
                    coord = coords[0] if coords else None
                    out[facet] = infer_cmor_like_time_frequency(
                        dset, rule.var or coord
                    )

    def extract_from_data(self, dset: xarray.Dataset) -> Dict[str, Any]:
        """Extract metadata from the data."""
        data: Dict[str, Any] = {}
        self._set_global_attributes(dset, data)
        self._set_variable_attributes(dset, data)
        self._apply_stats_rules(dset, data)
        return data


class Datasets(BaseModel):
    """Definition of datasets that should be crawled."""

    __pydantic_extra__: Dict[str, Any] = Field(init=False)
    model_config = ConfigDict(extra="allow")
    root_path: str | Path
    drs_format: str = "freva"
    fs_type: str = "posix"
    defaults: Dict[str, Any] = Field(default_factory=dict)
    storage_options: Dict[str, Any] = Field(default_factory=dict)
    glob_pattern: str = "*.*"
    inherits_from: str = Field(default_factory=str)

    @field_validator("storage_options", mode="after")
    @classmethod
    def _render_storage_options(
        cls, storage_options: Dict[str, Any]
    ) -> Dict[str, Any]:
        tmpl = TemplateMixin()
        return cast(Dict[str, Any], tmpl.render_templates(storage_options, {}))

    def model_post_init(self, __context: Any = None) -> None:
        """Apply rules after init."""
        storage_plugins = load_plugins("metadata_crawler.storage")
        try:
            self.backend = storage_plugins[self.fs_type](**self.storage_options)
        except KeyError:
            raise NotImplementedError(
                f"Backend not available. `{self.fs_type}` extension missing?"
            ) from None


class ConditionalRule(BaseModel):
    """Define conditional rules."""

    type: Literal["conditional"] = "conditional"
    condition: str
    true: Any
    false: Any


class CallRule(BaseModel):
    """Define caller rules."""

    type: Literal["call"] = "call"
    call: str


class LookupRule(BaseModel):
    """Define lookup table rules."""

    type: Literal["lookup"] = "lookup"
    tree: List[str] = Field(default_factory=list)
    attribute: str
    standard: Optional[str] = None


SpecialRule = Annotated[
    Union[ConditionalRule, CallRule, LookupRule], Field(discriminator="type")
]


class Dialect(BaseModel):
    """Settings for a DRS Format."""

    facets: Dict[str, str | list[str]] = Field(default_factory=dict)
    defaults: Dict[str, Any] = Field(default_factory=dict)
    path_specs: PathSpecs = Field(default_factory=PathSpecs)
    data_specs: DataSpecs = Field(default_factory=DataSpecs)
    special: Dict[str, SpecialRule] = Field(default_factory=dict)
    domains: Dict[str, List[float]] = Field(default_factory=dict)
    sources: List[MetadataSource] = Field(
        default_factory=lambda: [
            MetadataSource.path,
        ],
        description="Priority list of where to retrieve metadata",
    )
    inherits_from: Optional[str] = None

    @field_validator("sources", mode="after")
    @classmethod
    def _validate_sources(cls, srcs: List[str]) -> List[str]:
        # ensure only allowed sources are present
        names = {name.upper() for name in MetadataSource.__members__.keys()}
        values = {m.value for m in MetadataSource}
        invalid = [s for s in srcs if s.upper() not in names and s not in values]
        if invalid:
            allowed = sorted(values | {n.lower() for n in names})
            raise MetadataCrawlerException(
                f"Invalid metadata source(s): {invalid!r}. Allowed: {allowed}"
            )
        return srcs


class DRSConfig(BaseModel, TemplateMixin):
    """BaseModel model for the entire user config."""

    datasets: Dict[str, Datasets]
    index_schema: Dict[str, SchemaField] = Field(...)
    suffixes: List[str] = Field(default_factory=list)
    storage_options: Dict[str, Any] = Field(default_factory=dict)
    defaults: Dict[str, Any] = Field(default_factory=dict)
    special: Dict[str, SpecialRule] = Field(default_factory=dict)
    dialect: Dict[str, Dialect]

    def model_post_init(self, __context: Any = None) -> None:
        """Apply special rules after init."""
        self._defaults: Dict[str, Any] = {}
        self.suffixes = self.suffixes or [
            ".zarr",
            ".zar",
            ".nc4",
            ".nc",
            ".tar",
            ".hdf5",
            ".h5",
        ]
        for key, dset in self.datasets.items():
            self.dialect.setdefault(key, self.dialect[dset.drs_format])
            dset.backend.suffixes = self.suffixes
            for key, option in self.storage_options.items():
                dset.backend.storage_options.setdefault(key, option)
        for key, dset in self.datasets.items():
            self._defaults.setdefault(key, {})
            for k, _def in (dset.defaults or {}).items():
                self._defaults[key].setdefault(k, _def)
            for k, _def in self.dialect[dset.drs_format].defaults.items():
                self._defaults[key].setdefault(k, _def)
            for k, _def in self.defaults.items():
                self._defaults[key].setdefault(k, _def)
        self.prep_template_env()
        for standard in self.dialect:
            for key in self.special:
                self.dialect[standard].special.setdefault(key, self.special[key])

    @model_validator(mode="before")
    @classmethod
    def _dump_(cls, values: Any) -> Any:
        setattr(cls, "_model_dict", values)
        return values

    @model_validator(mode="before")
    def _resolve_inheritance(cls, values: Any) -> Any:
        """Apply inheritance.

        After loading raw TOML into dicts, but before model instantiation, merge
        any dialects that declare `inherits_from`.
        """
        if not isinstance(values, dict):
            return values  # pragma: no cover

        def _deep_merge(a: Dict[str, Any], b: Dict[str, Any]) -> None:
            for k, v in b.items():
                if k in a and isinstance(a[k], dict) and isinstance(v, dict):
                    if not v:
                        a[k] = {}
                    else:
                        _deep_merge(a[k], v)
                else:
                    a[k] = v

        for key in ("dialect", "datasets"):
            raw = values.get(key, {})
            merged = deepcopy(raw)
            for name, cfg in raw.items():
                parent = cfg.get("inherits_from")
                if parent:
                    if parent not in merged:
                        raise MetadataCrawlerException(
                            f"'{name}' inherits from unknown " f"'{parent}'"
                        )
                    # take parent base, then overlay this dialect
                    base = deepcopy(
                        merged[parent]
                    )  # shallow copy of parent raw dict
                    # remove inherits_from to avoid cycles
                    child = deepcopy(cfg)
                    child.pop("inherits_from", None)
                    # deep-merge child into base
                    _deep_merge(base, child)
                    base["inherits_from"] = parent
                    merged[name] = base

            values[key] = merged
        return values

    @model_validator(mode="before")
    def _ensure_dialects(cls, values: Any) -> Any:
        """Ensure every dialect is a Dialect model."""
        if not isinstance(values, dict):
            return values  # pragma: no cover

        raw = values.get("dialect", {})
        values["dialect"] = {k: v for k, v in raw.items()}
        return values

    def _apply_special_rules(
        self,
        standard: str,
        drs_type: str,
        inp: Metadata,
        specials: Dict[str, SpecialRule],
    ) -> None:
        data = {**inp.metadata, **{"file": inp.path, "uri": inp.path}}

        for facet, rule in specials.items():
            result: Any = None
            if inp.metadata.get(facet):
                continue
            match rule.type:
                case "conditional":
                    _rule = textwrap.dedent(rule.condition or "").strip()
                    s_cond = self.render_templates(_rule, data)
                    cond = eval(
                        s_cond, {}, getattr(self, "_model_dict", {})
                    )  # nosec
                    result = rule.true if cond else rule.false
                case "lookup":
                    args = cast(List[str], self.render_templates(rule.tree, data))

                    result = self.datasets[standard].backend.lookup(
                        inp.path,
                        self.render_templates(rule.attribute, data),
                        rule.standard or drs_type,
                        *args,
                        **self.dialect[standard].data_specs.read_kws,
                    )
                case "call":
                    _call = textwrap.dedent(rule.call or "").strip()
                    result = eval(
                        self.render_templates(_call, data),
                        {},
                        getattr(self, "_model_dict", {}),
                    )  # nosec
            if result:
                inp.metadata[facet] = result

    def _metadata_from_path(self, path: str, standard: str) -> Dict[str, Any]:
        """Extract the metadata from the path."""
        drs_type = self.datasets[standard].drs_format
        root_path = strip_protocol(
            self.datasets[standard].backend.path(
                self.datasets[standard].root_path
            )
        )
        _path = strip_protocol(self.datasets[standard].backend.path(path))
        rel_path = _path.with_suffix("").relative_to(root_path)
        return self.dialect[drs_type].path_specs.get_metadata_from_path(rel_path)

    @classmethod
    def load(
        cls,
        *config_paths: Union[Path, str, Dict[str, Any], tomlkit.TOMLDocument],
    ) -> DRSConfig:
        """Load a drs config from file."""
        cfg = tomli.loads(ConfigMerger(*config_paths).dumps())
        settings = cfg.pop("drs_settings")
        try:
            return cls(datasets=cfg, **settings)
        except ValidationError as e:
            msgs = []
            for err in e.errors():
                loc = ".".join(str(x) for x in err["loc"])
                msgs.append(f"{loc}: {err['msg']}")
            raise MetadataCrawlerException(
                "DRSConfig validation failed:\n" + "\n".join(msgs)
            ) from None

    def max_directory_tree_level(
        self, search_dir: str | Path, drs_type: str
    ) -> Tuple[int, bool]:
        """Get the maximum level for descending into directories.

        When searching for files in a directory we can only traverse the directory
        search tree until the version level is reached. This level is set as a hard
        threshold. If the drs type has no version we can indeed go all the way down
        to the file level.
        """
        root_path = strip_protocol(
            self.datasets[drs_type].backend.path(
                self.datasets[drs_type].root_path
            )
        )
        search_dir = strip_protocol(
            self.datasets[drs_type].backend.path(search_dir)
        )
        standard = self.datasets[drs_type].drs_format
        version = cast(
            str, self.dialect[standard].facets.get("version", "version")
        )
        is_versioned = True
        dir_parts = self.dialect[standard].path_specs.dir_parts or []
        try:
            version_idx = dir_parts.index(version)
        except ValueError:
            # No version given
            version_idx = len(dir_parts)
            is_versioned = False
        if root_path == search_dir:
            current_pos = 0
        else:
            current_pos = len(search_dir.relative_to(root_path).parts)
        return version_idx - current_pos, is_versioned

    def is_complete(self, data: Dict[str, Any], standard: str) -> bool:
        """Check if all metadata that can be collected was collected."""
        if not data:
            return False
        complete = True
        preset = {**self._defaults[standard], **self.dialect[standard].special}
        facets = (
            k for k, v in self.index_schema.items() if not v.key.startswith("__")
        )
        for facet in self.dialect[standard].facets or facets:
            if facet not in data and facet not in preset:
                complete = False
        return complete

    def _read_metadata(self, standard: str, inp: Metadata) -> Dict[str, Any]:
        """Get the metadata from a store."""
        drs_type = self.datasets[standard].drs_format
        for source in self.dialect[drs_type].sources:
            if self.is_complete(inp.metadata, standard) is True:
                break
            match source:
                case MetadataSource.path:
                    inp.metadata.update(
                        self._metadata_from_path(inp.path, standard)
                    )
                case MetadataSource.data:
                    with catch_warnings(action="ignore", category=RuntimeWarning):
                        with self.datasets[standard].backend.open_dataset(
                            inp.path, **self.dialect[standard].data_specs.read_kws
                        ) as ds:
                            inp.metadata.update(
                                self.dialect[
                                    standard
                                ].data_specs.extract_from_data(ds)
                            )
        self._apply_special_rules(
            standard, drs_type, inp, self.dialect[standard].special
        )
        return self._translate(standard, inp)

    def read_metadata(self, standard: str, inp: MetadataType) -> Dict[str, Any]:
        """Get the meta data for a given file path."""
        return self._read_metadata(
            standard,
            Metadata(path=inp["path"], metadata=inp["metadata"].copy()),
        )

    def _translate(self, standard: str, inp: Metadata) -> Dict[str, Any]:
        out: Dict[str, Any] = {}
        # locals to cut attribute lookups
        defs = self._defaults[standard]
        dia = self.dialect[standard]
        facets_get = dia.facets.get
        backend = self.datasets[standard].backend
        mget = inp.metadata.get
        defs_get = defs.get
        path = inp.path
        fmt = path.rsplit(".", 1)[1] if "." in path else ""

        precomputed: Dict[str, Any] = {
            "path": backend.path(path),
            "uri": backend.uri(path),
            "storage": backend.fs_type(path),
            "dataset": standard,
            "fmt": fmt,
        }
        val: Any = ""
        out_set = out.__setitem__
        for field, schema in self.index_schema.items():
            if schema.indexed is False:
                continue

            stype = schema.type

            # Fast path for simple, precomputed types
            if stype in precomputed and stype != "daterange":
                val = precomputed[stype]

            elif stype == "daterange":
                src = mget(field) or defs_get(field)
                val = schema.get_time_range(src)

            else:
                # Resolve metadata key via facets once; default to field name
                key = cast(str, facets_get(schema.key, field))
                val = mget(key) or defs_get(key)

            # Preserve your current semantics: fall back to schema.default on falsey
            val = val or schema.default

            # Multi-valued normalization
            if (
                (schema.multi_valued or schema.length)
                and val
                and not isinstance(val, list)
            ):
                val = [val]

            out_set(field, val)
        return out



"""Utilities to convert time delta to CMOR time frequencies."""

from __future__ import annotations

import datetime as dt
from typing import Any, Optional, cast

import numpy as np
import pandas as pd
import xarray as xr


def _seconds_from_timedelta(delta: Any) -> float:
    """Normalize timedelta-like objects to seconds (float)."""
    if hasattr(delta, "to_numpy"):
        delta = delta.to_numpy()

    if isinstance(delta, np.timedelta64):
        return cast(float, delta.astype("timedelta64[ns]").astype("int64") / 1e9)
    if isinstance(delta, dt.timedelta):
        return delta.total_seconds()
    if isinstance(delta, pd.Timedelta):
        return cast(float, delta.total_seconds())

    try:
        return float(delta)
    except Exception as exc:  # pragma: no cover
        raise TypeError(f"Unrecognized timedelta type: {type(delta)}") from exc


def _near(
    value: float, target: float, rel: float = 0.01, abs_tol: float = 60.0
) -> bool:
    """Compare two float values within relative or absolute tolerance."""
    return abs(value - target) <= max(abs_tol, abs(target) * rel)


def _map_seconds_to_cmor_like_frequency(sec: float) -> str:
    """
    Map a timestep in seconds to a CMOR-like frequency string.

    Returns one of:
        subhr, 1hr, 3hr, 6hr,
        day, 6d, 1w, sem,
        mon, season, yr, dec,
        <Xd> (generic fallback),
        unknown
    """
    day = 86400.0
    hour = 3600.0

    if sec <= 0:
        return "unknown"

    # Sub-hourly
    if sec < 0.5 * hour:
        return "subhr"

    # Hourly
    if _near(sec, hour):
        return "1hr"
    if _near(sec, 3 * hour):
        return "3hr"
    if _near(sec, 6 * hour):
        return "6hr"

    # Daily
    if _near(sec, day):
        return "day"

    # Multi-day (requested extras)
    if _near(sec, 6 * day):
        return "6d"
    if _near(sec, 7 * day):
        return "1w"
    if _near(sec, 14 * day):
        return "sem"

    # Monthly-ish
    if 20 * day <= sec <= 40 * day:
        return "mon"

    # Seasonal-ish (~3 months)
    if 80 * day <= sec <= 100 * day:
        return "season"

    # Yearly-ish
    if 350 * day <= sec <= 380 * day:
        return "yr"

    # Decadal-ish
    if 9 * 365 * day <= sec <= 11 * 365 * day:
        return "dec"

    # Fallback: express as days
    return f"{sec / day:.3g}d"


def _find_time_coord(
    ds: xr.Dataset,
    time_coord: Optional[str] = None,
) -> Optional[xr.DataArray]:
    """Best-effort detection of the time coordinate."""
    # 1) Explicit
    if time_coord is not None:
        if time_coord in ds.coords:
            return ds.coords[time_coord]
        if time_coord in ds.variables:
            return ds[time_coord]
        return None

    # 2) Coordinate literally named "time"
    if "time" in ds.coords:
        return ds.coords["time"]

    # 3) Coordinate with standard_name="time" or axis="T"
    for coord in ds.coords.values():
        std_name = coord.attrs.get("standard_name", "").lower()
        axis = coord.attrs.get("axis", "")
        if std_name == "time" or axis == "T":
            return cast(xr.DataArray, coord)

    # 4) Any coord whose dim is named "time"
    time_like_coords: list[xr.DataArray] = []
    for coord in ds.coords.values():
        if any(dim.lower() == "time" for dim in coord.dims):
            time_like_coords.append(coord)
    if time_like_coords:
        return time_like_coords[0]

    # 5) As last resort: any variable (coord or not) that looks time-like
    for vname in ds.variables:
        var = ds[vname]
        if any(str(dim).lower() == "time" for dim in var.dims):
            # Require datetime-like or object (for cftime) to avoid bogus matches
            if np.issubdtype(var.dtype, np.datetime64) or var.dtype == "O":
                return ds[vname]

    return None


def infer_cmor_like_time_frequency(
    ds: xr.Dataset,
    time_coord: Optional[str] = None,
) -> str:
    """
    Infer a CMOR-like time frequency from the first two valid time entries.

    Parameters
    ----------
    ds:
        Open xarray Dataset.
    time_coord:
        Optional explicit name of the time coordinate.

    Returns
    -------
    freq : str
        One of:
          - 'fx'        : no/insufficient/constant time
          - 'subhr'
          - '1hr', '3hr', '6hr'
          - 'day'
          - '6d', '1w', 'sem'
          - 'mon', 'season', 'yr', 'dec'
          - '<Xd>'      : generic days fallback
          - 'unknown'   : invalid/negative step
    """
    t = _find_time_coord(ds, time_coord=time_coord)

    if t is None:
        return "fx"

    # Ensure 1D along its primary dim
    if t.ndim != 1:
        main_dim = t.dims[0]
        t = t.isel({main_dim: slice(None)})

    if t.size < 2:
        return "fx"

    # Extract values
    vals = np.asarray(t.values).ravel()

    # Try via pandas for datetime64 / cftime that can be coerced
    dt_like = pd.to_datetime(vals, errors="coerce")
    valid = dt_like[~dt_like.isna()]

    if valid.size >= 2:
        uniq = np.unique(valid)
        if uniq.size < 2:
            return "fx"
        uniq.sort()
        delta = uniq[1] - uniq[0]
    else:
        # Fallback for non-coercible types: keep non-null, sort, diff
        non_null = [v for v in vals if v is not None]
        if len(non_null) < 2:
            return "fx"
        non_null = sorted(non_null)
        if non_null[0] == non_null[1]:
            return "fx"
        delta = non_null[1] - non_null[0]

    sec = _seconds_from_timedelta(delta)
    freq = _map_seconds_to_cmor_like_frequency(sec)
    return freq



"""Command line interface for the data crawler."""

from __future__ import annotations

import argparse
import asyncio
import inspect
import os
import sys
from functools import partial
from json import dumps
from pathlib import Path
from typing import (
    Annotated,
    Any,
    Callable,
    Dict,
    List,
    Optional,
    Tuple,
    Union,
    cast,
    get_args,
    get_origin,
    get_type_hints,
)

from rich_argparse import ArgumentDefaultsRichHelpFormatter

from metadata_crawler import add, delete, get_config, index

from ._version import __version__
from .api.metadata_stores import CatalogueBackends, IndexName
from .backends.intake import IntakePath
from .logger import (
    THIS_NAME,
    apply_verbosity,
    logger,
)
from .utils import exception_handler, load_plugins

StorageScalar = Union[str, int, float, bool]
StorageOptions = Dict[str, StorageScalar]
KwargValue = Union[
    str, int, float, Path, StorageOptions, List[str], List[int], None
]


def walk_catalogue(
    path: str,
    storage_options: Optional[Dict[str, Any]] = None,
    **kwargs: Any,
) -> int:
    """Recursively traverse an intake catalogue.

    Parameters
    ^^^^^^^^^^

    path:
        The path to the intake catalogue
    storage_options:
        Optional configuration passed to open catalogues residing on non posix
        storage backends, such as S3/MinIO
    """

    async def _walk(path: str, **storage_options: Any) -> int:
        num_items = 0
        ip = IntakePath(**storage_options)
        async for md in ip.walk(path):
            print(md)
            num_items += 1
        return num_items

    storage_options = storage_options or {}
    return asyncio.run(_walk(path, **storage_options))


def _flatten(inp: Union[List[str], List[List[str]]]) -> List[str]:

    out = []
    for item in inp:
        out += item if isinstance(item, list) else [item]
    return out


def _process_storage_option(option: str) -> Union[str, bool, int, float]:

    if option.lower() in ("false", "true"):
        return option.lower() == "true"
    try:
        return int(option)
    except ValueError:
        pass
    try:
        return float(option)
    except ValueError:
        pass
    return option


def display_config(
    *config: Union[Path, str], json: bool = False, **kwargs: Any
) -> None:
    """Display the config file."""
    cfg = get_config(*config)
    if json is False:
        print(cfg.dumps())
    else:
        print(dumps(cfg.merged_doc, indent=3))


class ArgParse:
    """Command line interface definition.

    Properties
    ----------
    kwargs: dict[str, Union[str, float, Path]]
            property holding all parsed keyword arguments.
    """

    kwargs: Optional[Dict[str, KwargValue]] = None
    verbose: int = 0
    epilog: str = (
        "See also "
        "https://metadata-crawler.readthedocs.io"
        " for a detailed documentation."
    )

    def __init__(self) -> None:
        """Instantiate the CLI class."""
        self.verbose: int = 0
        self.parser = argparse.ArgumentParser(
            prog=THIS_NAME,
            description="Add/Remove metadata to/from a metadata index.",
            formatter_class=ArgumentDefaultsRichHelpFormatter,
            epilog=self.epilog,
        )
        self.parser.add_argument(
            "-V",
            "--version",
            action="version",
            version=f"%(prog)s {__version__}",
            help="Print the version end exit",
        )
        self._add_general_config_to_parser(self.parser)
        self.subparsers = self.parser.add_subparsers(
            description="Collect or ingest metadata",
            required=True,
        )
        self._add_config_parser()
        self._add_walk_catalogue()
        self._add_crawler_subcommand()
        self._index_submcommands()

    def _add_config_parser(self) -> None:
        parser = self.subparsers.add_parser(
            "config",
            description="Display config",
            help="Display config",
            formatter_class=ArgumentDefaultsRichHelpFormatter,
            epilog=self.epilog,
        )
        parser.add_argument(
            "-c",
            "--config",
            help="Path(s) to the config_file(s)",
            type=Path,
            action="append",
        )
        parser.add_argument(
            "--json", help="Print in json format.", action="store_true"
        )
        parser.set_defaults(apply_func=display_config)
        parser.add_argument(
            "-v",
            "--verbose",
            action="count",
            default=self.verbose,
            help="Increase the verbosity level.",
        )

    def _add_crawler_subcommand(self) -> None:
        """Add sub command for crawling metadata."""
        parser = self.subparsers.add_parser(
            "add",
            description="Harvest (add) metadata",
            help="Harvest (crawl) metadata",
            formatter_class=ArgumentDefaultsRichHelpFormatter,
            epilog=self.epilog,
        )
        parser.add_argument(
            "store",
            type=str,
            help="Path to the intake catalogue",
        )
        parser.add_argument(
            "--catalogue-backend",
            "-cb",
            type=str,
            help="Source type of the catalogue backend.",
            choices=CatalogueBackends.__members__.keys(),
            default=list(CatalogueBackends.__members__.keys())[0],
        )
        parser.add_argument(
            "--data-store-prefix",
            "--prefix",
            type=str,
            help=(
                "Set the path prefix for the metadata store, this can either be"
                " an absolute path or if absolute path  is given a path prefix"
                " relative to the yaml catalogue file."
            ),
            default="metadata",
        )
        parser.add_argument(
            "-c",
            "--config-file",
            "--config-dir",
            type=Path,
            help="Path(s) to the metadata config file(s)",
            action="append",
            default=(os.environ.get("EVALUATION_SYSTEM_CONFIG_DIR"),),
        )
        parser.add_argument(
            "-b",
            "--batch-size",
            type=int,
            default=25_000,
            help="Set the batch size for ingestion.",
        )
        parser.add_argument(
            "--scan-concurrency",
            "--concurrency",
            type=int,
            default=1024,
            help="Level of aync concurrency for data discovery.",
        )
        parser.add_argument(
            "-d",
            "--data-object",
            "--data-obj",
            type=str,
            help="Objects (directories or catalogue files) that are processed.",
            default=None,
            action="append",
        ),
        parser.add_argument(
            "-ds",
            "--data-set",
            type=str,
            help=(
                "The name of the dataset(s) that are processed. "
                "names can contain wildcards such as ``xces-*``."
            ),
            default=None,
            action="append",
        )
        parser.add_argument(
            "-p",
            "--password",
            help=(
                "Ask for a password and set it to the DRS_STORAGE_PASSWD "
                "env variable."
            ),
            action="store_true",
        )
        parser.add_argument(
            "--n-procs",
            "--procs",
            help="Set the number of parallel processes for collecting.",
            type=int,
            default=None,
        )
        parser.add_argument(
            "--latest-version",
            type=str,
            default=IndexName().latest,
            help="Name of the core holding 'latest' metadata.",
        )
        parser.add_argument(
            "--all-versions",
            type=str,
            default=IndexName().all,
            help="Name of the core holding 'all' metadata versions.",
        )
        parser.add_argument(
            "--comp-level",
            "-z",
            help="Set the compression level for compressing files.",
            default=4,
            type=int,
        )
        parser.add_argument(
            "--storage_option",
            "-s",
            help=(
                "Set additional storage options for adding metadata to the"
                "metadata store"
            ),
            action="append",
            nargs=2,
        )
        parser.add_argument(
            "--fail-under",
            help=" Fail if less than X of the discovered files could be indexed.",
            type=int,
            default=-1,
        )
        parser.add_argument(
            "--shadow",
            help=(
                "'Shadow' these storage options. This is useful to hide secrets "
                "in public data catalogues."
            ),
            action="append",
            default=None,
            nargs="+",
        )
        self._add_general_config_to_parser(parser)
        parser.set_defaults(apply_func=add)

    def _add_general_config_to_parser(
        self, parser: argparse.ArgumentParser
    ) -> None:
        """Add the most common arguments to a given parser."""
        parser.add_argument(
            "-v",
            "--verbose",
            action="count",
            default=self.verbose,
            help="Increase the verbosity level.",
        )
        parser.add_argument(
            "--log-suffix",
            type=str,
            help="Add a suffix to the log file output.",
            default=None,
        )

    def _add_walk_catalogue(self) -> None:
        """Add a subcommand for walking an intake catalogue."""
        parser = self.subparsers.add_parser(
            "walk-intake",
            description="Walk an intake catalogue",
            help="Walk an intake catalogue",
            formatter_class=ArgumentDefaultsRichHelpFormatter,
            epilog=self.epilog,
        )

        parser.add_argument(
            "path",
            type=str,
            help="Path/Url to the intake catalogue",
        )
        parser.add_argument(
            "--storage_option",
            "-s",
            help=(
                "Set additional storage options for adding metadata to the"
                "metadata store"
            ),
            action="append",
            nargs=2,
        )
        parser.add_argument(
            "-v",
            "--verbose",
            action="count",
            default=self.verbose,
            help="Increase the verbosity level.",
        )
        parser.set_defaults(apply_func=walk_catalogue)

    def _index_submcommands(self) -> None:
        """Add sub command for adding metadata to the solr server."""
        entry_point = "metadata_crawler.ingester"
        for plugin, cls in load_plugins(entry_point).items():
            cli_methods: Dict[str, Callable[..., Any]] = {}
            for name in ("index", "delete"):
                method = getattr(cls, name, None)
                if hasattr(method, "_cli_help"):
                    cli_methods[name] = cast(Callable[..., Any], method)
            if cli_methods:
                subparser = self.subparsers.add_parser(
                    plugin,
                    help=cls.__doc__,
                    description=cls.__doc__,
                    formatter_class=ArgumentDefaultsRichHelpFormatter,
                    epilog=self.epilog,
                )
                cmd_parser = subparser.add_subparsers(required=True)
            for name, method in cli_methods.items():
                parser = cmd_parser.add_parser(
                    name,
                    help=getattr(method, "_cli_help", ""),
                    description=getattr(method, "_cli_help", ""),
                    formatter_class=ArgumentDefaultsRichHelpFormatter,
                    epilog=self.epilog,
                )
                parser.add_argument(
                    "-b",
                    "--batch-size",
                    type=int,
                    default=5_000,
                    help="Set the batch size for indexing.",
                )
                parser.add_argument(
                    "--storage_option",
                    "-s",
                    help=(
                        "Set additional storage options for adding metadata to "
                        " the metadata store"
                    ),
                    action="append",
                    nargs=2,
                )
                params = inspect.signature(method).parameters
                annotations = get_type_hints(method, include_extras=True)
                for param_name, param in params.items():
                    if param_name == "self":
                        continue
                    cli_meta = None
                    base_type = None
                    ann = annotations.get(param_name, None)

                    # 1) Annotated[...] style
                    if get_origin(ann) is Annotated:
                        base_type, *extras = get_args(ann)
                        # find the dict emitted by cli_parameter()
                        cli_meta = next(
                            (
                                e
                                for e in extras
                                if isinstance(e, dict) and "args" in e
                            ),
                            None,
                        )

                    # 2) default-as-parameter style
                    if (
                        cli_meta is None
                        and isinstance(param.default, dict)
                        and "args" in param.default
                    ):
                        cli_meta = param.default
                        # annotation is the base type
                        base_type = ann if ann is not inspect._empty else None

                    # if we found a cli_meta, wire it up
                    if cli_meta:
                        arg_names = cli_meta["args"]
                        add_kwargs = {
                            k: v for k, v in cli_meta.items() if k != "args"
                        }

                        # preserve any explicit default
                        if (
                            param.default is not inspect._empty
                            and cli_meta is not param.default
                        ):
                            add_kwargs["default"] = param.default

                        # enforce the base type if supplied
                        if base_type and "type" not in add_kwargs:
                            add_kwargs["type"] = base_type
                        parser.add_argument(
                            *arg_names, dest=param_name, **add_kwargs
                        )
                if name == "index":
                    parser.add_argument(
                        "catalogue_files",
                        help="File path to the metadata store.",
                        type=str,
                        nargs="*",
                    )

                    parser.set_defaults(
                        apply_func=partial(index, index_system=plugin)
                    )
                else:
                    parser.set_defaults(
                        apply_func=partial(delete, index_system=plugin)
                    )
                self._add_general_config_to_parser(parser)

    def parse_args(self, argv: list[str]) -> argparse.Namespace:
        """Parse the arguments for the command line interface.

        Parameters
        ----------
        argv: list[str]
            List of command line arguments that is parsed.

        Returns
        -------
        argparse.Namespace
        """
        args = self.parser.parse_args(argv)
        self.kwargs = {
            k: v
            for (k, v) in args._get_kwargs()
            if k
            not in (
                "apply_func",
                "verbose",
                "version",
                "storage_option",
                "shadow",
            )
        }
        storage_option_pairs: List[Tuple[str, str]] = (
            getattr(args, "storage_option", None) or []
        )
        so: StorageOptions = {}
        for option, value in storage_option_pairs:
            so[option] = _process_storage_option(value)
        if getattr(args, "shadow", None) or []:
            self.kwargs["shadow"] = _flatten(args.shadow)
        self.kwargs["storage_options"] = so
        self.verbose = args.verbose
        self.kwargs["verbosity"] = self.verbose
        return args


def _run(
    parser: argparse.Namespace,
    **kwargs: KwargValue,
) -> None:
    """Apply the parsed method."""
    old_level = apply_verbosity(
        getattr(parser, "verbose", 0), suffix=getattr(parser, "log_suffix", None)
    )
    cfg_files = cast(
        Tuple[Path, ...], kwargs.pop("config_file", kwargs.pop("config", ()))
    )
    try:
        parser.apply_func(*cfg_files, **kwargs)
    except Exception as error:
        exception_handler(error)
    finally:
        logger.set_level(old_level)


def cli(sys_args: list[str] | None = None) -> None:
    """Methods that creates the command line argument parser."""
    try:
        parser = ArgParse()
        args = parser.parse_args(sys_args or sys.argv[1:])
        kwargs = parser.kwargs or {}
        _run(args, **kwargs)
    except KeyboardInterrupt:
        raise SystemExit("Exiting program")



"""Collection of aync data ingest classes."""

from __future__ import annotations

import asyncio
import re
from functools import cached_property
from typing import Annotated, Any, Dict, List, Optional, Tuple
from urllib.parse import ParseResult, parse_qs, urlencode, urlparse, urlunparse

from motor.motor_asyncio import (
    AsyncIOMotorClient,
    AsyncIOMotorCollection,
    AsyncIOMotorDatabase,
)
from pymongo import DeleteMany, UpdateOne

from ..api.cli import cli_function, cli_parameter
from ..api.index import BaseIndex
from ..logger import logger


class MongoIndex(BaseIndex):
    """Ingest metadata into a mongoDB server."""

    def __post_init__(self) -> None:
        self._raw_uri = ""
        self._url = ""
        self._client: Optional[AsyncIOMotorClient[Any]] = None

    @property
    def uri(self) -> str:
        """Create the connection uri for the mongoDB."""
        if self._url:
            return self._url
        parsed_url = urlparse(self._raw_uri)
        query = parse_qs(parsed_url.query)
        if "timeout" not in parsed_url.query.lower():
            query["timeoutMS"] = ["5000"]
        new_query = urlencode(query, doseq=True)
        self._url = urlunparse(
            ParseResult(
                parsed_url.scheme or "mongodb",
                parsed_url.netloc,
                parsed_url.path.rstrip("/"),
                parsed_url.params,
                new_query,
                parsed_url.fragment,
            )
        )
        return self._url

    @cached_property
    def unique_index(self) -> str:
        """Get the index."""
        for name, schema in self.index_schema.items():
            if schema.unique:
                return name
        raise ValueError("The schema doesn't define a unique value.")

    @property
    def client(self) -> AsyncIOMotorClient[Any]:
        """Get the mongoDB client."""
        if self._client is None:
            logger.debug("Creating async mongoDB client: %s", self.uri)
            self._client = AsyncIOMotorClient(self.uri)
        return self._client

    async def _bulk_upsert(
        self, chunk: List[Dict[str, Any]], collection: AsyncIOMotorCollection[Any]
    ) -> None:
        ops = [
            UpdateOne(
                {self.unique_index: m[self.unique_index]},
                {"$set": m},
                upsert=True,
            )
            for m in chunk
        ]
        await collection.bulk_write(ops, ordered=False)

    async def _index_collection(
        self, db: AsyncIOMotorDatabase[Any], collection: str, suffix: str = ""
    ) -> None:
        """Index a collection."""
        col = collection + suffix
        await db[col].create_index(self.unique_index, unique=True)
        async for chunk in self.get_metadata(collection):
            await self._bulk_upsert(chunk, db[col])

    async def _prep_db_connection(
        self, database: str, url: str
    ) -> AsyncIOMotorDatabase[Any]:

        await self.close()
        self._raw_uri = url or ""
        return self.client[database]

    @cli_function(
        help="Add metadata to the mongoDB metadata server.",
    )
    async def index(
        self,
        *,
        url: Annotated[
            Optional[str],
            cli_parameter(
                "--url",
                help="The <host>:<port> to the mngoDB server",
                type=str,
            ),
        ] = None,
        database: Annotated[
            str,
            cli_parameter(
                "--database",
                "--db",
                help="The DB name holding the metadata.",
                type=str,
                default="metadata",
            ),
        ] = "metadata",
        index_suffix: Annotated[
            Optional[str],
            cli_parameter(
                "--index-suffix",
                help="Suffix for the latest and all version collections.",
                type=str,
            ),
        ] = None,
    ) -> None:
        """Add metadata to the mongoDB metadata server."""
        db = await self._prep_db_connection(database, url or "")
        async with asyncio.TaskGroup() as tg:
            for collection in self.index_names:
                tg.create_task(
                    self._index_collection(
                        db, collection, suffix=index_suffix or ""
                    )
                )

    async def close(self) -> None:
        """Close the mongoDB connection."""
        self._client.close() if self._client is not None else None
        self._url = ""
        self._raw_uri = ""

    @cli_function(
        help="Remove metadata from the mongoDB metadata server.",
    )
    async def delete(
        self,
        *,
        url: Annotated[
            Optional[str],
            cli_parameter(
                "--url",
                help="The <host>:<port> to the mngoDB server",
                type=str,
            ),
        ] = None,
        database: Annotated[
            str,
            cli_parameter(
                "--database",
                "--db",
                help="The DB name holding the metadata.",
                type=str,
                default="metadata",
            ),
        ] = "metadata",
        facets: Annotated[
            Optional[List[Tuple[str, str]]],
            cli_parameter(
                "-f",
                "--facets",
                type=str,
                nargs=2,
                action="append",
                help="Search facets matching the delete query.",
            ),
        ] = None,
    ) -> None:
        """Remove metadata from the mongoDB metadata server."""
        db = await self._prep_db_connection(database, url or "")
        if not facets:
            logger.info("Nothing to delete")
            return

        def glob_to_regex(glob: str) -> str:
            """Turn a shell‐style glob into a anchored mongo regex."""
            # escape everything, then un-escape our wildcards
            esc = re.escape(glob)
            esc = esc.replace(r"\*", ".*").replace(r"\?", ".")
            return f"^{esc}$"

        ops: List[DeleteMany] = []
        for field, val in facets:
            if "*" in val or "?" in val:
                pattern = glob_to_regex(val)
                ops.append(DeleteMany({field: {"$regex": pattern}}))
            else:
                ops.append(DeleteMany({field: val}))
        logger.debug("Deleting entries matching %s", ops)
        for collection in await db.list_collection_names():
            await db[collection].bulk_write(ops, ordered=False)



"""Test crawling swift stores."""

from pathlib import Path

import intake
import pytest

from metadata_crawler import add
from metadata_crawler.backends.swift import SwiftPath


def test_crawl_obs(
    drs_config_path: Path,
    cat_file: Path,
) -> None:
    """Test crawling swift obs."""

    add(
        drs_config_path,
        store=cat_file,
        batch_size=3,
        n_procs=1,
        data_set=["obs-swift"],
        catalogue_backend="jsonlines",
    )
    assert cat_file.exists()
    cat = intake.open_catalog(cat_file)
    assert len(cat.latest.read()) > 0


def test_crawl_cmip6(
    drs_config_path: Path,
    cat_file: Path,
) -> None:

    add(
        drs_config_path,
        store=cat_file,
        batch_size=3,
        n_procs=1,
        data_set=["cmip6-swift"],
    )
    assert cat_file.exists()
    cat = intake.open_catalog(cat_file)
    assert len(cat.latest.read()) > 0
    # There are versioned datasets so latest should not have all the entries
    assert len(cat.latest.read()) < len(cat.files.read())


def test_crawl_single_file(
    drs_config_path: Path,
    cat_file: Path,
) -> None:
    """Test if we can crawl a single file on s3."""
    file = (
        "http://localhost:8081/v1/AUTH_test/test/model/global/cmip6/CMIP6/CMIP"
        "/CSIRO-ARCCSS/ACCESS-CM2/amip/r1i1p1f1/Amon/ua/gn/v20191108/"
        "ua_Amon_ACCESS-CM2_amip_r1i1p1f1_gn_197001-201512.nc"
    )
    add(
        drs_config_path,
        store=cat_file,
        n_procs=1,
        batch_size=3,
        data_object=[file],
    )
    assert cat_file.is_file()
    cat = intake.open_catalog(cat_file)
    assert len(cat.latest.read()) == len(cat.files.read()) == 1


def test_crawl_zarr(drs_config_path: Path, cat_file: Path) -> None:
    """Zarr stores are directories but are treated as files."""

    add(
        drs_config_path,
        store=cat_file,
        data_set=["nextgems-swift"],
        n_procs=1,
        batch_size=3,
        verbosity=4,
    )
    assert cat_file.is_file()
    cat = intake.open_catalog(cat_file)
    assert len(cat.latest.read()) > 0


@pytest.mark.asyncio
async def test_crawl_no_auth(drs_config_path: Path) -> None:
    """No file available."""
    p = SwiftPath(
        os_storage_url="http://localhost:8081/v1/AUTH_test/test",
    )
    assert p.os_auth_url
    with pytest.raises(ValueError):
        await p.is_file("/foo")
    p = SwiftPath()
    assert p.os_auth_url == ""
    with pytest.raises(RuntimeError):
        await p.is_dir("/foo")
    p = SwiftPath(
        os_storage_url="http://localhost:8081/v1/AUTH_test/test",
        os_password="testing",
        os_project_id="test2",
        os_user_id="tester",
    )
    num = 0
    with pytest.raises(ValueError):
        async for file in p.iterdir("/model/nextgems"):
            num += 1
    assert num == 0


@pytest.mark.asyncio
async def test_crawl_no_file(drs_config_path: Path) -> None:
    """No auth available."""
    p = SwiftPath(
        os_storage_url="http://localhost:8081/v1/AUTH_test/test",
        os_password="testing",
        os_project_id="test",
        os_user_id="tester",
    )
    num = 0
    async for file in p.iterdir("/foo/bar"):
        num += 1
    assert num == 0
